package de.uka.ipd.sdq.dsexplore.helper;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;
import org.eclipse.core.runtime.CoreException;
import org.opt4j.core.Value;

import de.uka.ipd.sdq.dsexplore.exception.ExceptionHelper;
import de.uka.ipd.sdq.dsexplore.opt4j.genotype.DesignDecisionGenotype;
import de.uka.ipd.sdq.dsexplore.opt4j.representation.DSEDecoder;
import de.uka.ipd.sdq.dsexplore.opt4j.representation.DSEObjectives;
import de.uka.ipd.sdq.dsexplore.opt4j.representation.DSEProblem;
import de.uka.ipd.sdq.dsexplore.opt4j.start.Opt4JStarter;
import de.uka.ipd.sdq.pcm.designdecision.Choice;
import de.uka.ipd.sdq.pcm.designdecision.DegreeOfFreedom;
import de.uka.ipd.sdq.statistics.estimation.ConfidenceInterval;

public class GenotypeReader {
	
	private static final String SEPARATOR = ";";
	
	/** Logger for log4j. */
	private static Logger logger = 
		Logger.getLogger("de.uka.ipd.sdq.dsexplore.helper.GenotypeReader");

	/**
	 * Reads the genotypes from the specified files. 
	 * If the file contains pretty printed genotypes and 
	 * objectives are given, reads them and updates the DSEEvaluator cache
	 * by calling DSEEvaluator.addToPhenotypeCache(..). Requires confidence intervals
	 * to be available for all objectives with statistical evaluators. 
	 * Reading in of objectives and confidence values is not supported for raw genotype files. 
	 */
	public static List<DesignDecisionGenotype> getGenotypes(String filename) throws CoreException{
		File file = new File(filename);
		
	    try {
	    	
	    	List<DesignDecisionGenotype> results = readInPrettyPrintedGenotypes(getReaderFor(file));
	        
	        return results;
	        
	      } catch( Exception ex ) {
	        throw ExceptionHelper.createNewCoreException("Could not evaluate predefined instances. Leave field blank for evolutionary search or type \"random\" for a random search. "+ex.getMessage(), ex);
	      }
	}	

	private static BufferedReader getReaderFor(File file) throws FileNotFoundException {
    	InputStreamReader ir = new InputStreamReader(
                new FileInputStream( file ) );
        BufferedReader in = new BufferedReader(ir);
		return in;
	}

	/**
	 * 
	 * @param line
	 * @return
	 */
	private static boolean isLineOfDoublesOnly(String line) {
		boolean isDoubleLine = true;
		String[] lineArray = line.split(SEPARATOR);
		
		//try all values in the line
    	try {
        	for (String string : lineArray) {
				Double.parseDouble(string);
			}
    	} catch (NumberFormatException e){
    		isDoubleLine = false;
    	}
		return isDoubleLine;
	}

	/**
	 * Reads the genotypes from the specified files. 
	 * If objectives are given, reads them and updates the DSEEvaluator cache
	 * by calling DSEEvaluator.addToPhenotypeCache(..). Requires confidence intervals
	 * to be available for all objectives with statistical evaluators. 
	 * @param in Is used and modify and closed 
	 * @return
	 * @throws CoreException
	 * @throws IOException
	 */
	private static List<DesignDecisionGenotype> readInPrettyPrintedGenotypes(BufferedReader in) throws CoreException, IOException {
		List<DesignDecisionGenotype> results = new ArrayList<DesignDecisionGenotype>();
		
		DSEProblem problem = Opt4JStarter.getProblem();
		List<DegreeOfFreedom> decisionList = problem.getDesignDecisions();
		
		//order design decisions
		String headline = in.readLine();
		if (headline == null){
			throw ExceptionHelper.createNewCoreException("Predefined instances file could not be read because it is empty. Specify a valid file or \"random\" or leave the field blank for an evolutionary search.");
		} 
		String[] headlineArray = headline.split(SEPARATOR);
		
		List<ObjectiveAndEvaluator> objectiveList = Opt4JStarter.getDSEEvaluator().getObjectiveAndEvaluatorList();
		
		//Empty list if no objectives are specified. 
		List<ObjectiveAndEvaluator> orderedObjectives = getOrderedObjectiveCollection(headlineArray, objectiveList);
		boolean hasObjectives = orderedObjectives.size() != 0;
		
		//adjust evaluators.
		if (orderedObjectives.size() != 0 && objectiveList.size() > orderedObjectives.size()){
			Opt4JStarter.getDSEEvaluator().retainOnlyEvaluatorsFor(orderedObjectives);
		} 
		
		int startIndexOfConfidence = orderedObjectives.size();
		
		//now count confidence intervals. 
		int numberOfConfidenceIntervals = 0;
		for (ObjectiveAndEvaluator objectiveAndEvaluator : orderedObjectives) {
			if (objectiveAndEvaluator.getEvaluator().hasStatisticResults()){
				numberOfConfidenceIntervals++;
			}
		}
		
		//Its always three columns per objective that has confidence
		int startIndexOfGenotype =  orderedObjectives.size() + numberOfConfidenceIntervals*3;
		
		List<DegreeOfFreedom> orderedDesignDecisions = new ArrayList<DegreeOfFreedom>();
		for (int i = startIndexOfGenotype; i < headlineArray.length; i++) {
			String headlineEntry = headlineArray[i];
			for (DegreeOfFreedom decision : decisionList) {
				//We could use an iterator here and directly delete all design decision that have been matched from the list, to make the subsequent steps faster.
				if (DegreeOfFreedomHelper.getDegreeDescription(decision).equals(headlineEntry)){
					orderedDesignDecisions.add(decision);
					break;
				}
			}
		}
		//set the internal design decisions to the same order
		problem.getDesignDecisions().clear();
		problem.getDesignDecisions().addAll(orderedDesignDecisions);
		
		int expectedLineArrayLength = startIndexOfGenotype + orderedDesignDecisions.size();
		
		String line;
		while ( null != (line = in.readLine()) ) {
			String[] lineArray = line.split(SEPARATOR);
			//Check minimum length, a last semicolon is ok that makes the line array one longer than the list.
			if (lineArray.length < expectedLineArrayLength){
				throw ExceptionHelper.createNewCoreException("A line in the predefined instances file has not the right length");
			}
			DesignDecisionGenotype genotype = new DesignDecisionGenotype();

			//read in genotype
			int decisionIndex = 0;
			for (int i = startIndexOfGenotype; i < lineArray.length && decisionIndex < orderedDesignDecisions.size(); i++) {
				Choice gene = DSEDecoder.getChoiceFor(lineArray[i], orderedDesignDecisions.get(decisionIndex));
				genotype.add(gene);
				decisionIndex++;
			}
			results.add(genotype);

			if (hasObjectives){
				//Read in objectives
				DSEObjectives objectives = readInObjectives(lineArray, startIndexOfConfidence, orderedObjectives);

				//read in confidence
				readInConfidenceIntervals(lineArray, startIndexOfConfidence, numberOfConfidenceIntervals, orderedObjectives, objectives);
			
				//add objective to phenotype cache with genotype string
				Opt4JStarter.getDSEEvaluator().addToPhenotypeCache(DSEDecoder.getGenotypeString(genotype), objectives);
			}
			
		}
		in.close();
		
		return results;
	}

	/**
	 * Fills the confidence intervals of the passed objective and returns it. 
	 * @param lineArray
	 * @param startIndexOfConfidence
	 * @param numberOfConfidenceIntervals
	 * @param orderedObjectiveList
	 * @param objectives
	 * @return
	 * @throws CoreException
	 */
	private static DSEObjectives readInConfidenceIntervals(String[] lineArray,
			int startIndexOfConfidence, int numberOfConfidenceIntervals,
			List<ObjectiveAndEvaluator> orderedObjectiveList, DSEObjectives objectives) throws CoreException {
		
		//read in confidence 
		int index = startIndexOfConfidence;
		for (ObjectiveAndEvaluator objectiveAndEvaluator : orderedObjectiveList) {
			if (objectiveAndEvaluator.getEvaluator().hasStatisticResults() && index < lineArray.length - 2){
				String lowerConfidenceString = lineArray[index];
				String upperConfidenceString = lineArray[index+1];
				String alphaConfidenceString = lineArray[index+2];
				ConfidenceInterval ci = readInConfidenceInterval(lowerConfidenceString, upperConfidenceString, alphaConfidenceString, objectives.get(objectiveAndEvaluator.getObjective()));
				index = index +3;
				objectives.addConfidence(objectiveAndEvaluator.getObjective(), ci);
			} else {
				
			}
		}
		return objectives;
	}
	
	/**
	 * Objective definition must be in the first columns of the file. 
	 * @param headline
	 * @param objectiveList 
	 * @return never null, but may be an empty list.
	 */
	private static List<ObjectiveAndEvaluator> getOrderedObjectiveCollection(String[] headline, List<ObjectiveAndEvaluator> objectiveList) throws CoreException{
		
		List<ObjectiveAndEvaluator> orderedObjectives = new ArrayList<ObjectiveAndEvaluator>(objectiveList.size());
		
		for (int i = 0; i < headline.length; i++) {
			String objectiveString = headline[i];
			ObjectiveAndEvaluator o = matchObjective(objectiveList, objectiveString);
			if (o == null){
				//The current entry is not an objective, so the objective columns are done
				break;
			} else {
				orderedObjectives.add(o);
			}
		}
		
		if (objectiveList.size() < orderedObjectives.size()){
			// I cannot have more objectives than evaluators. 
			// throw Exception and notify user that there must be only one column per objective
			throw ExceptionHelper.createNewCoreException("Error reading in objectives: There must be only one column for each objective.");
		}
	
		return orderedObjectives;
	}
	
	
	private static ConfidenceInterval readInConfidenceInterval(
			String lowerConfidenceString, String upperConfidenceString,
			String alphaConfidenceString,
			Value<?> value) throws CoreException {
		
		try {
			double lowerBound = parseToDouble(lowerConfidenceString);
			double upperBound = parseToDouble(upperConfidenceString);
			double level = parseToDouble(alphaConfidenceString);
			return new ConfidenceInterval(value.getDouble(), lowerBound, upperBound, level);
		} catch (NumberFormatException e){
			throw ExceptionHelper.createNewCoreException("Could not read in confidence interval. Expected this to be a confidence interval: "
					+lowerConfidenceString+SEPARATOR+upperConfidenceString+SEPARATOR+alphaConfidenceString, e);
		}
		
		
	}
	
	
	
	private static ObjectiveAndEvaluator matchObjective(
			List<ObjectiveAndEvaluator> objectiveList, String objectiveString) {
		for (ObjectiveAndEvaluator objective : objectiveList) {
			if (objectiveString.equals(objective.getObjective().toString()) || objectiveString.equals(objective.getObjective().getName())){
				return objective;
			}
		}
		return null;
	}
	
	/**
	 * Reads in 
	 * @param orderedObjectives 
	 * @param endIndex 
	 * @param 
	 * @return
	 * @throws IOException
	 * @throws CoreException 
	 */
	private static DSEObjectives readInObjectives(String[] objectiveLine, int endIndex, List<ObjectiveAndEvaluator> orderedObjectives) throws CoreException {
		
		//objective line must have at least the length of the passed list
		if (objectiveLine.length < orderedObjectives.size()){
			throw ExceptionHelper.createNewCoreException("Error when reading in result line: the line is shorter ("+objectiveLine+" entries) than the number of objectives I look for ("+orderedObjectives.size()+"entries).");
		}
		
		DSEObjectives obj = new DSEObjectives();
		int lineIndex = 0;
		for (ObjectiveAndEvaluator oe : orderedObjectives) {
			try {
				double d = parseToDouble(objectiveLine[lineIndex]);
				obj.add(oe.getObjective(),d);
			} catch (NumberFormatException e){
				throw ExceptionHelper.createNewCoreException("Could not parse objective value "+objectiveLine[lineIndex]+". Fix your input genome file.");
			}
			lineIndex++;
		}
		
		return obj;
	}

	private static double parseToDouble(String number) {
		double d =  Double.parseDouble(number);
		if (Double.isInfinite(d)){
			logger.warn("Read in an infinite objective or confidence interval from predefined file, please check that this is correct and not result of an incomplete previous analysis step.");
		} else if (Double.isNaN(d)){
			logger.warn("Read in an objective or confidence interval with NaN value from predefined file, please check that this is correct and not result of an incomplete previous analysis step.");
		}
		return d;
	}

}
