package de.uka.ipd.sdq.dsexplore.launch;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.apache.log4j.Logger;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.debug.core.ILaunch;
import org.opt4j.core.DoubleValue;
import org.opt4j.core.Value;

import de.uka.ipd.sdq.dsexplore.analysis.AnalysisProxy;
import de.uka.ipd.sdq.dsexplore.analysis.IAnalysis;
import de.uka.ipd.sdq.dsexplore.helper.GenotypeReader;
import de.uka.ipd.sdq.dsexplore.opt4j.genotype.DesignDecisionGenotype;
import de.uka.ipd.sdq.dsexplore.opt4j.start.GivenInstanceEvaluator;
import de.uka.ipd.sdq.dsexplore.opt4j.start.Opt4JStarter;
import de.uka.ipd.sdq.dsexplore.opt4j.start.PredefinedInstanceEvaluator;
import de.uka.ipd.sdq.dsexplore.opt4j.start.RandomSearch;
import de.uka.ipd.sdq.pcmsolver.models.PCMInstance;
import de.uka.ipd.sdq.workflow.IBlackboardInteractingJob;
import de.uka.ipd.sdq.workflow.IJob;
import de.uka.ipd.sdq.workflow.exceptions.JobFailedException;
import de.uka.ipd.sdq.workflow.exceptions.RollbackFailedException;
import de.uka.ipd.sdq.workflow.exceptions.UserCanceledException;
import de.uka.ipd.sdq.workflow.mdsd.blackboard.MDSDBlackboard;
import de.uka.ipd.sdq.workflow.pcm.blackboard.PCMResourceSetPartition;
import de.uka.ipd.sdq.workflow.pcm.jobs.LoadPCMModelsIntoBlackboardJob;

public class OptimisationJob implements IJob, IBlackboardInteractingJob<MDSDBlackboard> {
	
	/** Logger for log4j. */
	private static Logger logger = 
		Logger.getLogger("de.uka.ipd.sdq.dsexplore.launch.OptimisationJob");

	private PCMInstance pcmInstance;
	private List<IAnalysis> evaluators;
	private ArrayList<Value<Double>> upperConstraints;
	private Long startTimestampMillis;
	private ILaunch launch;
	
	private MDSDBlackboard blackboard;
	
	//TODO: Quickfix to reset the Loggers. Refactor!
	private DSELaunch dseLaunch;
	private DSEWorkflowConfiguration dseConfig;
    	

	public OptimisationJob(DSEWorkflowConfiguration config, DSELaunch launch) {
		
		    this.dseLaunch = launch; 
		    this.dseConfig = config;
		    this.launch = dseLaunch.getOriginalLaunch();
	}



	@Override
	public void execute(IProgressMonitor monitor) throws JobFailedException,
			UserCanceledException {
		
		startTimestampMillis = System.currentTimeMillis();
		
		logger.debug("Starting...");

		try{
		logger.debug("\n Launching optimization with \n"+
				"maximal number of iterations: "+this.dseConfig.getMaxIterations()+"\n"+
				"individuals per generation:" +this.dseConfig.getIndividualsPerGeneration()+"\n"+
				"offspring per generation: individuals per generation / 2 rounded up.");
		
		pcmInstance = new PCMInstance((PCMResourceSetPartition)this.blackboard.getPartition(LoadPCMModelsIntoBlackboardJob.PCM_MODELS_PARTITION_ID));
	    List<PCMInstance> instances = new ArrayList<PCMInstance>();
	    instances.add(pcmInstance);
	    
	    //initialise the algorithm and analysis
	    //HillClimbingAlgorithm algorithm = new HillClimbingAlgorithm(mrtRequirements);
	    //IAlgorithm algorithm = new FullSearchAlgorithm();
		
	    upperConstraints = new ArrayList<Value<Double>>();
		
	    evaluators = new ArrayList<IAnalysis>();
	    if (dseConfig.isAnalysisEnabledFor(DSEConstantsContainer.PERFORMANCE)){
	    	evaluators.add(new AnalysisProxy(this.blackboard, this.dseConfig.getRawConfiguration(), launch, monitor, DSEConstantsContainer.PERFORMANCE, this.dseLaunch));
		upperConstraints.add(new DoubleValue(this.dseConfig.getMrtRequirements()));
	    }
	    if (dseConfig.isAnalysisEnabledFor(DSEConstantsContainer.RELIABILITY)){
	    	evaluators.add(new AnalysisProxy(this.blackboard, this.dseConfig.getRawConfiguration(), launch, monitor, DSEConstantsContainer.RELIABILITY, this.dseLaunch));
	    	upperConstraints.add(new DoubleValue(this.dseConfig.getMaxPOFOD()));
	    }
	    if (dseConfig.isAnalysisEnabledFor(DSEConstantsContainer.COST)){
	    	evaluators.add(new AnalysisProxy(this.blackboard, this.dseConfig.getRawConfiguration(), launch, monitor, DSEConstantsContainer.COST, this.dseLaunch));
	    	upperConstraints.add(new DoubleValue(this.dseConfig.getMaxCost()));
	    }
		
	    //algorithm.initialise(instances, perfAnalysisTool,configuration);

	    Opt4JStarter.init(evaluators, upperConstraints,this.dseConfig,getPCMInstance());
	    
	    if (this.dseConfig.hasCacheInstances()){
	    	fillCacheWithValues(this.dseConfig.getCacheInstancesFileName());
	    }
	    
		//TODO: extract this in a Builder?
	    if (this.dseConfig.isOptimise()){
	    	if (!dseConfig.hasPredefinedInstances()){
	    		Opt4JStarter.startOpt4J(evaluators, this.dseConfig, upperConstraints, monitor,getPCMInstance());
	    	} else if (dseConfig.isRandomSearch()){
	    		PredefinedInstanceEvaluator rie = new RandomSearch(this.dseConfig);
	    		rie.start();
	    	} else if (this.dseConfig.getMaxIterations() == 0){
	    		//Only evaluate given instances
	    		PredefinedInstanceEvaluator gie = new GivenInstanceEvaluator(this.dseConfig);
	    		gie.start();

	    	} else {
	    		//use predefined instances as initial population
	    		String populationFilename = this.dseConfig.getPredefinedInstancesFileName();
	    		List<DesignDecisionGenotype> genotypes = GenotypeReader.getGenotypes(populationFilename);
	    		
	    		//read in all candidates file if there and add to archive
	    		String allCandidatesFilename = this.dseConfig.getPredefinedAllCandidatesFileName();
	    		List<DesignDecisionGenotype> allCandidates;
	    		if (!"".equals(allCandidatesFilename)){
	    			allCandidates = GenotypeReader.getGenotypes(allCandidatesFilename);
	    		} else {
	    			allCandidates = Collections.emptyList();
	    		}
	    		
	    		Opt4JStarter.runOpt4JWithPopulation(this.dseConfig, monitor, genotypes, allCandidates);
	    	}
	    }
		} catch (CoreException e){
			throw new JobFailedException(e);
	  		
		} finally {
		
			logger.warn("DSE launch done. It took "+((System.currentTimeMillis()-startTimestampMillis)/1000)+" seconds.");
			
			try {
				/* There was quite some memory allocation at this point (probably),
				 * which seemed caused an OutOfMemory quicker than the VM could get 
				 * more memory (max perm size had not been reached yet). Thus, make sure
				 * here that all memory is freed. */ 
				Runtime.getRuntime().gc();
				Opt4JStarter.closeTask();
				Opt4JStarter.tearDown();
				Runtime.getRuntime().gc();
			} catch (Exception e){
				e.printStackTrace();
			}

		}

	}


	private void fillCacheWithValues(String cacheInstancesFileName) throws CoreException {
		GenotypeReader.getGenotypes(cacheInstancesFileName);
	}



	@Override
	public String getName() {
		return this.getClass().getName();
	}

	@Override
	public void rollback(IProgressMonitor monitor)
			throws RollbackFailedException {
		// TODO Auto-generated method stub

	}



	@Override
	public void setBlackboard(MDSDBlackboard blackboard) {
		this.blackboard = blackboard;
		
	}
	
	private PCMInstance getPCMInstance(){
		if (this.blackboard != null){
			return new PCMInstance((PCMResourceSetPartition)this.blackboard.getPartition(LoadPCMModelsIntoBlackboardJob.PCM_MODELS_PARTITION_ID));
		} else {
			String message = "Internal error: Cannot retrieve PCM model if blackboard is not set. ";
			logger.error(message);
			throw new UnsupportedOperationException(message);
		}
	}
	
}
