package de.uka.ipd.sdq.dsexplore.opt4j.operator;



import org.opt4j.genotype.DoubleGenotype;
import org.opt4j.genotype.IntegerGenotype;
import org.opt4j.operator.mutate.Mutate;
import org.opt4j.operator.mutate.MutateDouble;
import org.opt4j.operator.mutate.MutateInteger;

import com.google.inject.Inject;

import de.uka.ipd.sdq.dsexplore.exception.ChoiceOutOfBoundsException;
import de.uka.ipd.sdq.dsexplore.exception.InvalidChoiceForDegreeException;
import de.uka.ipd.sdq.dsexplore.opt4j.genotype.DesignDecisionGenotype;
import de.uka.ipd.sdq.pcm.designdecision.Choice;
import de.uka.ipd.sdq.pcm.designdecision.ContinousRangeChoice;
import de.uka.ipd.sdq.pcm.designdecision.ContinuousRangeDegree;
import de.uka.ipd.sdq.pcm.designdecision.DegreeOfFreedom;
import de.uka.ipd.sdq.pcm.designdecision.DiscreteRangeChoice;
import de.uka.ipd.sdq.pcm.designdecision.DiscreteRangeDegree;
import de.uka.ipd.sdq.pcm.designdecision.EnumerationChoice;
import de.uka.ipd.sdq.pcm.designdecision.EnumerationDegree;

/**
 * Mutator for {@link DesignDecisionGenotype}s. Calls the bound {@link MutateInteger} or {@link MutateDouble}, depending on 
 * the type of {@link DegreeOfFreedom} of each gene. 
 * 
 * @author martens
 *
 */
public class MutateDesignDecisionGenotype implements Mutate<DesignDecisionGenotype> {
	
	MutateInteger mutateInteger;
	MutateDouble mutateDouble;
	
	@Inject
	public MutateDesignDecisionGenotype(MutateInteger mutateInteger, MutateDouble mutateDouble){
		 this.mutateDouble = mutateDouble;
		 this.mutateInteger = mutateInteger;
	}

	/**
	 * Mutates the passed {@link DesignDecisionGenotype} by calling a matching {@link Mutate}
	 * for each gene.
	 */
	@Override
	public void mutate(DesignDecisionGenotype genotype) {
		for (Choice choice : genotype) {
			if (choice instanceof DiscreteRangeChoice){
				mutateDiscrete((DiscreteRangeChoice)choice);
			} else if (choice instanceof EnumerationChoice){
				mutateEnum((EnumerationChoice)choice);
			} else if (choice instanceof ContinousRangeChoice){
				mutateContinous((ContinousRangeChoice)choice);
			} else throw new UnsupportedOperationException("Choice type "+choice+" not supported.");
		}
		
	}

	/**
	 * Calls {@link MutateDouble} for the given {@link Choice}.
	 * @param choice
	 */
	private void mutateContinous(ContinousRangeChoice choice) {
		DegreeOfFreedom degree = choice.getDegreeOfFreedom();
		if (degree instanceof ContinuousRangeDegree){
			ContinuousRangeDegree contDegree = (ContinuousRangeDegree)degree;
			DoubleGenotype doubleList = new DoubleGenotype(contDegree.getFrom(),contDegree.getTo());
			doubleList.add(choice.getChosenValue());
			this.mutateDouble.mutate(doubleList);
			double newValue = doubleList.get(0);
			choice.setChosenValue(newValue);
		} else throw new InvalidChoiceForDegreeException(choice);
	}

	/**
	 * Calls {@link MutateInteger} for the given {@link Choice}.
	 * @param choice
	 */
	private void mutateEnum(EnumerationChoice choice) {
		DegreeOfFreedom degree = choice.getDegreeOfFreedom();
		if (degree instanceof EnumerationDegree){
			EnumerationDegree enumDegree = (EnumerationDegree) degree;
			int oldIndex = enumDegree.getDomainOfEntities().indexOf(choice.getEntity());
			if (oldIndex == -1){
				throw new ChoiceOutOfBoundsException(choice, "Error when mutating individual, old choice was invalid");
			}
			int newIndex = mutateInteger(oldIndex, 0, enumDegree.getDomainOfEntities().size()-1);
			if (newIndex < 0 || newIndex >= enumDegree.getDomainOfEntities().size()){
				throw new RuntimeException("Error when mutating integer index value: Value is out of bounds!");
			}
			choice.setEntity(enumDegree.getDomainOfEntities().get(newIndex));
		}
	}

	/**
	 * Calls {@link MutateInteger} for the given {@link Choice}. 
	 * @param choice
	 */
	private void mutateDiscrete(DiscreteRangeChoice choice) {
		DegreeOfFreedom degree = choice.getDegreeOfFreedom();
		if (degree instanceof DiscreteRangeDegree) {
			DiscreteRangeDegree discDegree = (DiscreteRangeDegree) degree;
			int newValue = mutateInteger(choice.getChosenValue(), discDegree.getFrom(), discDegree.getTo());
			choice.setChosenValue(newValue);
		} else throw new InvalidChoiceForDegreeException(choice);
		
	}

	private int mutateInteger(int oldValue, int lowerBound, int upperBound) {
		IntegerGenotype integerGenotype = new IntegerGenotype(lowerBound, upperBound);
		integerGenotype.add(oldValue);
		this.mutateInteger.mutate(integerGenotype);
		int newValue = integerGenotype.get(0);
		return newValue;
	}

}
