package de.uka.ipd.sdq.dsexplore.opt4j.optimizer.heuristic.operators;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.Random;

import org.opt4j.core.problem.Genotype;
import org.opt4j.operator.copy.Copy;

import de.uka.ipd.sdq.dsexplore.launch.DSEWorkflowConfiguration;
import de.uka.ipd.sdq.dsexplore.opt4j.representation.DSEIndividual;
import de.uka.ipd.sdq.dsexplore.opt4j.representation.DSEIndividualBuilder;
import de.uka.ipd.sdq.dsexplore.opt4j.start.Opt4JStarter;

public class HeuristicManager {
	/**
	 * Formating string used for logging purposes
	 */
	private static final String DATE_FORMAT_NOW = "yyyy-MM-dd HH:mm:ss";
	
	/**
	 * Heuristics to be used to generate candidates
	 */
	private Collection<IHeuristic> heuristics = new ArrayList<IHeuristic>();
	
	/**
	 * @param copy Creates copy of genotypes
	 * @param individualBuilder Builds individual
	 */
	public HeuristicManager(Copy<Genotype> copy, DSEIndividualBuilder individualBuilder) {
		DSEWorkflowConfiguration configuration = Opt4JStarter.getDSEWorkflowConfig();
		heuristics = HeuristicFactory.getActivatedInstances(copy, individualBuilder, configuration);
	}
	
	/**
	 * Applies registered heuristics to individual and returns candidates
	 * @param individual
	 * @return
	 */
	public Collection<HeuristicCandidate> getAllCandidates(DSEIndividual individual) {
		Collection<HeuristicCandidate> result = new ArrayList<HeuristicCandidate>();
		Collection<HeuristicCandidate> candidatesFromCurrentHeuristic;
		for (IHeuristic heuristic : heuristics) {
			candidatesFromCurrentHeuristic = heuristic.getHeuristicCandidates(individual);
			if (candidatesFromCurrentHeuristic.size() > 0) {
				    writeToLogFile(heuristic.getClass() + ";" + candidatesFromCurrentHeuristic.size() + "; candidate(s)");
			}
			result.addAll(candidatesFromCurrentHeuristic);
		}
		return result;
	}
	
	/**
	 * Applies registered heuristics to individual and returns random candidate
	 * based on candidate weight and heuristic weight. Returns null if no heuristic
	 * can be be applied.
	 * @param individual
	 * @return Random candidate from registered heuristic or null if no heuristic can be applied
	 */
	public DSEIndividual getCandidate(DSEIndividual individual) {
		Collection <HeuristicCandidate> candidates = getAllCandidates(individual);
		double sumOfAllWeights = 0;
		for (HeuristicCandidate c : candidates) {
			sumOfAllWeights += c.getFinalWeight();
		}
		Random generator = new Random();
		double threshold = generator.nextDouble()*sumOfAllWeights;
		double currentLevel = 0;
		for (HeuristicCandidate c : candidates) {
			currentLevel += c.getFinalWeight();
			// use spectrum of candidate weights to select candidate. The larger the
			// weight of the candidate the higher the probability of being selected
			if (currentLevel >= threshold) {
				writeToLogFile(c.getHeuristic().getClass() + ";candidate returned");
				return c;
			}
		}
		if (candidates.size() > 0) {
			throw new RuntimeException("Random selection failed.");
		} 
		// won't be executed unless candidates list is empty
		else {
			return null;
		}
	}
	
	public Collection<IHeuristic> getHeuristics() {
		return heuristics;
	}
	
	/**
	 * Writes String entry to log file in results directory. 
	 * @param entry
	 */
	private void writeToLogFile(String entry) {
		try{
		    // Create file 
			File decisionFileDir = (new File(Opt4JStarter.getDSEWorkflowConfig().getDesignDecisionFileName())).getParentFile(); 
			File resultsDir = new File(decisionFileDir, "PerOpteryx_results");
			File logFile = new File(resultsDir, "heuristicsInfo_" + Opt4JStarter.getDSEWorkflowConfig().startTimestamp + ".csv");
		    FileWriter fstream = new FileWriter(logFile.getAbsoluteFile(), true);
		    BufferedWriter out = new BufferedWriter(fstream);
		    Calendar cal = Calendar.getInstance();
		    SimpleDateFormat sdf = new SimpleDateFormat(DATE_FORMAT_NOW);
		    out.append(sdf.format(cal.getTime()) + ";" + entry);
		    out.newLine();
		    //Close the output stream
		    out.close();
		}catch (Exception e){//Catch exception if any
		      System.err.println("Error: " + e.getMessage());
		}
	}
	
}
