package de.uka.ipd.sdq.dsexplore.opt4j.optimizer.heuristic.operators.impl;

import java.util.ArrayList;
import java.util.Collection;

import org.opt4j.core.problem.Genotype;
import org.opt4j.operator.copy.Copy;

import de.uka.ipd.sdq.dsexplore.helper.UtilisationHelper;
import de.uka.ipd.sdq.dsexplore.launch.DSEWorkflowConfiguration;
import de.uka.ipd.sdq.dsexplore.opt4j.optimizer.heuristic.operators.AbstractHeuristic;
import de.uka.ipd.sdq.dsexplore.opt4j.optimizer.heuristic.operators.HeuristicCandidate;
import de.uka.ipd.sdq.dsexplore.opt4j.representation.DSEIndividual;
import de.uka.ipd.sdq.dsexplore.opt4j.representation.DSEIndividualBuilder;
import de.uka.ipd.sdq.dsexplore.opt4j.start.Opt4JStarter;
import de.uka.ipd.sdq.pcm.designdecision.Choice;
import de.uka.ipd.sdq.pcm.designdecision.ContinousRangeChoice;
import de.uka.ipd.sdq.pcm.designdecision.DegreeOfFreedom;
import de.uka.ipd.sdq.pcm.designdecision.ProcessingRateDegree;
import de.uka.ipd.sdq.pcm.resourceenvironment.ProcessingResourceSpecification;
import de.uka.ipd.sdq.pcm.resourcetype.ProcessingResourceType;
import de.uka.ipd.sdq.pcm.resultdecorator.resourceenvironmentdecorator.UtilisationResult;
import de.uka.ipd.sdq.pcm.resultdecorator.resourceenvironmentdecorator.impl.ProcessingResourceSpecificationResultImpl;

/**
 * This class implements an IHeuristic which increases the processing rate of
 * highly utilized processing resources and decreases the processing rate of
 * hardly utilized processing resources.
 * 
 * @author Tom Beyer
 */
public class ProcessingRateImpl extends AbstractHeuristic {
	/**
	 * Processing rate will be decreased by this factor if preconditions are
	 * fulfilled
	 */
	private double decreaseProcessingRateFactor;

	/**
	 * Processing rate will be increased by this factor if preconditions are
	 * fulfilled
	 */
	private double increaseProcessingRateFactor;

	/**
	 * If utilisation is larger than this double it will be considered a high
	 * utilisation
	 */
	private double thresholdHighUtilisation;
	/**
	 * If utilisation is smaller than this double it will be considered a low
	 * utilisation
	 */
	private double thresholdLowUtilisation;

	/**
	 * @param copy
	 *            Creates copy of genotypes
	 * @param individualBuilder
	 *            Builds individual
	 */
	public ProcessingRateImpl(Copy<Genotype> copy, DSEIndividualBuilder individualBuilder, DSEWorkflowConfiguration configuration) {
		super(copy, individualBuilder, configuration);
		// set config
		decreaseProcessingRateFactor = configuration.getProcessingRateDecreaseFactor();
		setHeuristicWeight(configuration.getProcessingRateWeight());
		increaseProcessingRateFactor = configuration.getProcessingRateIncreaseFactor();
		thresholdHighUtilisation = configuration.getProcessingRateThresholdHighUtilisation();
		thresholdLowUtilisation = configuration.getProcessingRateThresholdLowUtilisation();
	}

	/**
	 * Returns true if maximum utilisation is above or equals
	 * thresholdHighUtilisation and not null
	 * 
	 * @param individual
	 * @return
	 */
	private boolean doesMatchHighUtilisation(DSEIndividual individual) {
		UtilisationResult maxUtilisationResult = UtilisationHelper.getMaxUtilisationResult(individual);
		return maxUtilisationResult != null
				&& maxUtilisationResult.getNormalisedResourceUtilisation() >= thresholdHighUtilisation;
	}

	/**
	 * Returns true if maximum utilisation is below or equals
	 * thresholdLowUtilisation and not null
	 * 
	 * @param individual
	 * @return
	 */
	private boolean doesMatchLowUtilisation(DSEIndividual individual) {
		UtilisationResult minUtilisationResult = UtilisationHelper.getMinUtilisationResult(individual);
		return minUtilisationResult != null
				&& minUtilisationResult.getNormalisedResourceUtilisation() <= thresholdLowUtilisation;
	}

	/**
	 * Check whether individual matches precondition. Precondition: Does a
	 * resource exists which is either highly or hardly utilised
	 * 
	 * @param individual
	 *            Individual to check
	 */
	public boolean doesMatchPrecondition(DSEIndividual individual) {
		// Check whether there is one server with high or low utilization
		return doesMatchLowUtilisation(individual) || doesMatchHighUtilisation(individual);
	}

	/**
	 * Generates candidates based on given individual
	 * 
	 * @param Indivdual
	 *            used to apply heuristic
	 * @return Collection of generated candidates.
	 */
	public Collection<HeuristicCandidate> getHeuristicCandidates(DSEIndividual individual) {
		Collection<HeuristicCandidate> candidates = new ArrayList<HeuristicCandidate>(); // return value
		/*
		 * 1. Get maximum utilisation 2. Copy current genotype 3. Find
		 * processing resource by iterating through genotype and change
		 * processing rate 4. Add candidate to result collection
		 */
		if (doesMatchHighUtilisation(individual)) {
			addNewCandidateWithIncreasedProcessingRate(individual, candidates);
		}
		/*
		 * 1. Get minimum utilisation 2. Copy current genotype 3. Find
		 * processing resource by iterating through genotype and change
		 * processing rate 4. Add candidate to result collection
		 */
		if (doesMatchLowUtilisation(individual)) {
			addNewCandidateWithDecreasedProcessingRate(individual, candidates);
		}
		return candidates;
	}

	/**
	 * @param individual
	 * @param candidates
	 */
	private void addNewCandidateWithIncreasedProcessingRate(DSEIndividual individual,
			Collection<HeuristicCandidate> candidates) {
		// 1. Get maximum utilisation
		UtilisationResult maxUtilisationResult = UtilisationHelper.getMaxUtilisationResult(individual);
		ProcessingResourceType processingResourceType = getProcessingResourceTyp(maxUtilisationResult);
		// 2. Copy current genotype
		HeuristicCandidate candidate = individualBuilder.buildCandidate((copy.copy(individual.getGenotype())));
		// 3. Iterate through choices and find processing rate degree to change
		for (Choice choice : candidate.getGenotype()) {
			if (choice instanceof ContinousRangeChoice) {
				ContinousRangeChoice continousRangeChoice = (ContinousRangeChoice) choice;
				DegreeOfFreedom degreeOfFreedom = choice.getDegreeOfFreedom();
				if (degreeOfFreedom instanceof ProcessingRateDegree) {
					ProcessingRateDegree processingRateDegree = (ProcessingRateDegree) degreeOfFreedom;
					if (processingRateDegree.getProcessingresourcetype() == processingResourceType) {
						// actually increase processing rate. Respect
						// maximum allowed value of processing rate
						double newProcessingRate = getIncreasedProcessingRate(continousRangeChoice,
								processingRateDegree);
						continousRangeChoice.setChosenValue(newProcessingRate);

					}
				}
			}
		}
		candidate.setCandidateWeight(getCandidateWeightForHighUtilisation(maxUtilisationResult));
		candidate.setHeuristic(this);
		// 4. Add candidate to result collection
		candidates.add(candidate);
		increaseCounterOfGeneratedCandidates();
	}

	/**
	 * @param individual
	 * @param candidates
	 */
	private void addNewCandidateWithDecreasedProcessingRate(DSEIndividual individual,
			Collection<HeuristicCandidate> candidates) {
		// 1. Get maximum utilisation
		UtilisationResult minUtilisationResult = UtilisationHelper.getMinUtilisationResult(individual);
		ProcessingResourceType processingResourceType = getProcessingResourceTyp(minUtilisationResult);
		// 2. Copy current genotype
		HeuristicCandidate candidate = individualBuilder.buildCandidate(copy.copy(individual.getGenotype()));
		// 3. Iterate through choices and find processing rate degree to change
		for (Choice choice : candidate.getGenotype()) {
			if (choice instanceof ContinousRangeChoice) {
				ContinousRangeChoice continousRangeChoice = (ContinousRangeChoice) choice;
				DegreeOfFreedom degreeOfFreedom = choice.getDegreeOfFreedom();
				if (degreeOfFreedom instanceof ProcessingRateDegree) {
					ProcessingRateDegree processingRateDegree = (ProcessingRateDegree) degreeOfFreedom;
					if (processingRateDegree.getProcessingresourcetype() == processingResourceType) {
						// actually decrease processing rate. Respect
						// minimum allowed value of processing rate
						double newProcessingRate = getDecreasedProcessingRate(continousRangeChoice, processingRateDegree);
						continousRangeChoice.setChosenValue(newProcessingRate);

					}
				}
			}
		}
		candidate.setCandidateWeight(getCandidateWeightForLowUtilisation(minUtilisationResult));
		candidate.setHeuristic(this);
		candidates.add(candidate);
		increaseCounterOfGeneratedCandidates();
	}

	/**
	 * @param continousRangeChoice
	 * @param processingRateDegree
	 * @return
	 */
	private double getDecreasedProcessingRate(ContinousRangeChoice continousRangeChoice,
			ProcessingRateDegree processingRateDegree) {
		return Math.max(continousRangeChoice.getChosenValue()
				* (1 - decreaseProcessingRateFactor), processingRateDegree.getFrom());
	}

	/**
	 * @param continousRangeChoice
	 * @param processingRateDegree
	 * @return
	 */
	private double getIncreasedProcessingRate(ContinousRangeChoice continousRangeChoice,
			ProcessingRateDegree processingRateDegree) {
		return Math.min(continousRangeChoice.getChosenValue() * (1 + increaseProcessingRateFactor),
				processingRateDegree.getTo());
	}

	/**
	 * @param utilisationResult
	 * @return Processing Resource Type of utilisationResult
	 */
	private ProcessingResourceType getProcessingResourceTyp(UtilisationResult utilisationResult) {
		ProcessingResourceSpecification maxProcessingResourceSpec = ((ProcessingResourceSpecificationResultImpl) utilisationResult)
				.getProcessingresourcespecification();
		ProcessingResourceType processingResourceType = maxProcessingResourceSpec
				.getActiveResourceType_ActiveResourceSpecification();
		return processingResourceType;
	}

	/**
	 * Calculates priority based on the following scheme: if utilisation equals
	 * THRESHOLD_HIGH_UTLISATION then it will return 0, if utilisation equals 1
	 * it will return 1. Values in between are linearly extrapolated. Return
	 * values will always be >= 0.
	 * 
	 * @param utilisationResult
	 * @return Priority based on utilisationResult's utilisation
	 */
	private double getCandidateWeightForHighUtilisation(UtilisationResult utilisationResult) {
		return Math.max(utilisationResult.getNormalisedResourceUtilisation() / (1.0 - thresholdHighUtilisation)
				- thresholdHighUtilisation / (1.0 - thresholdHighUtilisation), 0);
	}

	/**
	 * Calculates weight based on the following scheme: if utilisation equals
	 * THRESHOLD_LOW_UTLISATION then it will return 0, if utilisation equals 0
	 * it will return 1. Values in between are linearly extrapolated. Return
	 * values will always be >= 0.
	 * 
	 * @param utilisationResult
	 * @return Weight based on utilisationResult's utilisation
	 */
	private double getCandidateWeightForLowUtilisation(UtilisationResult utilisationResult) {
		return Math.max(1.0 - utilisationResult.getNormalisedResourceUtilisation() / thresholdHighUtilisation, 0);
	}

}
