package de.uka.ipd.sdq.dsexplore.opt4j.optimizer.heuristic.operators.impl;

import java.util.ArrayList;
import java.util.Collection;

import org.opt4j.core.problem.Genotype;
import org.opt4j.operator.copy.Copy;

import de.uka.ipd.sdq.dsexplore.helper.UtilisationHelper;
import de.uka.ipd.sdq.dsexplore.launch.DSEWorkflowConfiguration;
import de.uka.ipd.sdq.dsexplore.opt4j.optimizer.heuristic.operators.AbstractHeuristic;
import de.uka.ipd.sdq.dsexplore.opt4j.optimizer.heuristic.operators.HeuristicCandidate;
import de.uka.ipd.sdq.dsexplore.opt4j.representation.DSEIndividual;
import de.uka.ipd.sdq.dsexplore.opt4j.representation.DSEIndividualBuilder;
import de.uka.ipd.sdq.dsexplore.opt4j.start.Opt4JStarter;
import de.uka.ipd.sdq.pcm.designdecision.AllocationDegree;
import de.uka.ipd.sdq.pcm.designdecision.Choice;
import de.uka.ipd.sdq.pcm.designdecision.EnumerationChoice;
import de.uka.ipd.sdq.pcm.resourceenvironment.ProcessingResourceSpecification;
import de.uka.ipd.sdq.pcm.resourceenvironment.ResourceContainer;
import de.uka.ipd.sdq.pcm.resultdecorator.resourceenvironmentdecorator.UtilisationResult;
import de.uka.ipd.sdq.pcm.resultdecorator.resourceenvironmentdecorator.impl.ProcessingResourceSpecificationResultImpl;

/**
 * Implements a reallocation heuristic which allocates a component from
 * a highly utilised resource container to a lowly utilised container
 * @author Tom Beyer
 *
 */
public class ReallocationImpl extends AbstractHeuristic {
	/**
	 * Maximum number of components to be reallocated
	 */
	private int maxNumberOfReplacements;
	/**
	 * If utilisation is larger than this double it will be considered
	 * a high utilisation
	 */
	private double thresholdHighUtilisation;
	/**
	 * If utilisation is smaller than this double it will be considered
	 * a low utilisation
	 */
	private double thresholdLowUtilisation;
	
	/**
	 * 
	 * @param copy Used to copy genotype
	 * @param individualBuilder Used to build individual
	 */
	public ReallocationImpl(Copy<Genotype> copy, DSEIndividualBuilder individualBuilder, DSEWorkflowConfiguration configuration) {
		super(copy, individualBuilder, configuration);
		// set config
		setHeuristicWeight(configuration.getReallocationWeight());
		maxNumberOfReplacements = configuration.getReallocationMaxNumberOfReplacements();
		thresholdHighUtilisation = configuration.getReallocationThresholdHighUtilisation();
		thresholdLowUtilisation = configuration.getReallocationThresholdLowUtilisation();
	}
	

	/**
	 * Check whether there is one server with high utilization
	 * and one with low utilization which are both not null
	 * @param individual Individual to check
	 */
	public boolean doesMatchPrecondition(DSEIndividual individual) {
		UtilisationResult minUtilisationResult = UtilisationHelper.getMinUtilisationResult(individual);
		UtilisationResult maxUtilisationResult = UtilisationHelper.getMaxUtilisationResult(individual);
		return minUtilisationResult != null && maxUtilisationResult != null && 
				maxUtilisationResult.getNormalisedResourceUtilisation() >= thresholdHighUtilisation &&
				minUtilisationResult.getNormalisedResourceUtilisation() <= thresholdLowUtilisation;
	}

	/**
	 * Generates collection of candidates by applying the reallocation heuristic
	 * @param individual Individual which the heuristic should be applied to
	 */
	public Collection<HeuristicCandidate> getHeuristicCandidates(DSEIndividual individual) {
		// return value
		Collection<HeuristicCandidate> candidates = new ArrayList<HeuristicCandidate>();
		if (doesMatchPrecondition(individual)){
			// create new candidate (find servers), set priority and add to list candidates
			UtilisationResult minUtilisationResult = UtilisationHelper.getMinUtilisationResult(individual);
			UtilisationResult maxUtilisationResult = UtilisationHelper.getMaxUtilisationResult(individual);
			// create candidate
			HeuristicCandidate candidate = individualBuilder.buildCandidate(copy.copy(individual.getGenotype()));
			ProcessingResourceSpecification minProcessingResourceSpec = ((ProcessingResourceSpecificationResultImpl)minUtilisationResult).getProcessingresourcespecification();
			ProcessingResourceSpecification maxProcessingResourceSpec = ((ProcessingResourceSpecificationResultImpl)maxUtilisationResult).getProcessingresourcespecification();
			ResourceContainer targetResourceContainer = ((ResourceContainer)minProcessingResourceSpec.eContainer());
			ResourceContainer sourceResourceContainer = ((ResourceContainer)maxProcessingResourceSpec.eContainer());
			if (targetResourceContainer == sourceResourceContainer) {
				return candidates;
			}
			int numberOfReplacements = 0;
			// iterate through choices and change AllocationDegree
			for (Choice choice : candidate.getGenotype()) {
				if (choice instanceof EnumerationChoice) {
					EnumerationChoice enumerationChoice = (EnumerationChoice)choice;
					if (enumerationChoice.getDegreeOfFreedom() instanceof AllocationDegree) {
						if (enumerationChoice.getEntity() == sourceResourceContainer && numberOfReplacements < maxNumberOfReplacements) {
							// deploy the first components you can find
							enumerationChoice.setEntity(targetResourceContainer);
							numberOfReplacements++;
						}
					}
				}
			}
			candidate.setCandidateWeight(getCandidateWeight(minUtilisationResult, maxUtilisationResult));
			candidate.setHeuristic(this);
			candidates.add(candidate);
			increaseCounterOfGeneratedCandidates();
		}
		return candidates;
	}
	
	/**
	 * Calculates weight of the candidate using maxUtil-minUtil
	 * @param minUtilisationResult UtilisationResult with minimum utilisation
	 * @param maxUtilisationResult UtilisationResult with maximum utilisation
	 * @return Weight of candidate
	 */
	private double getCandidateWeight(UtilisationResult minUtilisationResult,
			UtilisationResult maxUtilisationResult) {
		return maxUtilisationResult.getNormalisedResourceUtilisation() - minUtilisationResult.getNormalisedResourceUtilisation();
	}
}
