package de.uka.ipd.sdq.dsexplore.opt4j.optimizer.heuristic.operators.impl;

import java.util.ArrayList;
import java.util.Collection;

import org.opt4j.core.problem.Genotype;
import org.opt4j.operator.copy.Copy;

import de.uka.ipd.sdq.dsexplore.helper.UtilisationHelper;
import de.uka.ipd.sdq.dsexplore.launch.DSEWorkflowConfiguration;
import de.uka.ipd.sdq.dsexplore.opt4j.optimizer.heuristic.operators.AbstractHeuristic;
import de.uka.ipd.sdq.dsexplore.opt4j.optimizer.heuristic.operators.HeuristicCandidate;
import de.uka.ipd.sdq.dsexplore.opt4j.representation.DSEIndividual;
import de.uka.ipd.sdq.dsexplore.opt4j.representation.DSEIndividualBuilder;
import de.uka.ipd.sdq.dsexplore.opt4j.start.Opt4JStarter;
import de.uka.ipd.sdq.pcm.designdecision.AllocationDegree;
import de.uka.ipd.sdq.pcm.designdecision.Choice;
import de.uka.ipd.sdq.pcm.designdecision.EnumerationChoice;
import de.uka.ipd.sdq.pcm.resourceenvironment.ProcessingResourceSpecification;
import de.uka.ipd.sdq.pcm.resourceenvironment.ResourceContainer;
import de.uka.ipd.sdq.pcm.resultdecorator.resourceenvironmentdecorator.UtilisationResult;
import de.uka.ipd.sdq.pcm.resultdecorator.resourceenvironmentdecorator.impl.ProcessingResourceSpecificationResultImpl;

/**
 * Implements a server consolidation heuristic. This heuristic finds two
 * hardly utilised resource containers and reallocates all components of these
 * two resource containers to one resource container
 * @author Tom Beyer
 */
public class ServerConsolidationImpl extends AbstractHeuristic {
	/**
	 * If utilisation is smaller than this double it will be considered
	 * a low utilisation
	 */
	private double thresholdLowUtilisation;	

	/**
	 * 
	 * @param copy Used to copy genotype
	 * @param individualBuilder Used to build individual
	 */
	public ServerConsolidationImpl(Copy<Genotype> copy,
			DSEIndividualBuilder individualBuilder, DSEWorkflowConfiguration configuration) {
		super(copy, individualBuilder, configuration);
		// set config
		setHeuristicWeight(configuration.getServerConsolidationWeight());
		thresholdLowUtilisation = configuration.getServerConsolidationThresholdLowUtilisation();
		
	}
	
	/**
	 * Checks whether there are two resource containers with a 
	 * utilisation <= THRESHOLD_LOW_UTILISATION and which are both not null
	 * @param individual
	 * @return
	 */
	private boolean doesMatchLowUtilisation(DSEIndividual individual) {
		UtilisationResult minUtilisationResult = UtilisationHelper
			.getMinUtilisationResult(individual);
		UtilisationResult minUtilisationResult2 = UtilisationHelper.get2ndMinUtilisationResult(individual);
		return minUtilisationResult != null && 
			minUtilisationResult.getNormalisedResourceUtilisation() <= thresholdLowUtilisation &&
			minUtilisationResult2 != null &&
			minUtilisationResult2.getNormalisedResourceUtilisation() <= thresholdLowUtilisation;
	}
	
	/**
	 * Checks whether there are two resource containers with a 
	 * utilisation <= THRESHOLD_LOW_UTILISATION
	 * @param individual
	 * @return
	 */
	public boolean doesMatchPrecondition(DSEIndividual individual) {
		return doesMatchLowUtilisation(individual);
	}

	/**
	 * Generates collection of candidates by applying the server consolidation heuristic
	 * @param individual Individual which the heuristic should be applied to
	 */
	public Collection<HeuristicCandidate> getHeuristicCandidates(DSEIndividual individual) {
		Collection<HeuristicCandidate> candidates = new ArrayList<HeuristicCandidate>();
		if (doesMatchPrecondition(individual)) {
			/*
			 * 1. Find lowest and 2nd lowest utilised resource containers
			 * 2. Copy individual
			 * 3. Reallocate all components from one resource container
			 *    to the other
			 */
			// 1. Find lowest and 2nd lowest utilised resource containers
			UtilisationResult minUtilisationResult = UtilisationHelper.getMinUtilisationResult(individual);
			UtilisationResult minUtilisationResult2 = UtilisationHelper.get2ndMinUtilisationResult(individual);
			// 2. Copy individual
			HeuristicCandidate candidate = individualBuilder.buildCandidate(copy.copy(individual.getGenotype()));
			ProcessingResourceSpecification minProcessingResourceSpec = ((ProcessingResourceSpecificationResultImpl) minUtilisationResult)
					.getProcessingresourcespecification();
			ProcessingResourceSpecification minProcessingResourceSpec2 = ((ProcessingResourceSpecificationResultImpl) minUtilisationResult2)
					.getProcessingresourcespecification();
			ResourceContainer resourceContainer = ((ResourceContainer) minProcessingResourceSpec.eContainer());
			ResourceContainer resourceContainer2 = ((ResourceContainer) minProcessingResourceSpec2.eContainer());
			// 3. Reallocate all components from one resource container to the other
			for (Choice choice : candidate.getGenotype()) {
				if (choice instanceof EnumerationChoice) {
					EnumerationChoice enumerationChoice = (EnumerationChoice)choice;
					if (enumerationChoice.getDegreeOfFreedom() instanceof AllocationDegree) {
						if (enumerationChoice.getEntity() == resourceContainer) {
							// deploy all the components you can find to resourceContainer2
							enumerationChoice.setEntity(resourceContainer2);
						}
					}
				}
			}
			candidate.setCandidateWeight(getCandidateWeight(minUtilisationResult, minUtilisationResult2));
			candidate.setHeuristic(this);
			candidates.add(candidate);
			increaseCounterOfGeneratedCandidates();
		}
		return candidates;
	}
	
	/**
	 * Calculates weight of candidate using the following: 1-(utilisation1+utilisation2)/(2*THRESHOLD_LOW_UTLISATION).
	 * This means that the weight is 0 if both utilisations equal the threshold and 1 if they are both zero.
	 * In between linear interpolation is used. 
	 * @param u1 Lowest utilised resource container
	 * @param u2 2nd lowest utilised resource container
	 * @return Weight of candidate
	 */
	private double getCandidateWeight(UtilisationResult u1, UtilisationResult u2) {
		 return 1.0 - ((u1.getNormalisedResourceUtilisation() + u2.getNormalisedResourceUtilisation())/(2*thresholdLowUtilisation));
	}
	
}
