package de.uka.ipd.sdq.dsexplore.opt4j.optimizer.heuristic.operators.impl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Random;

import org.opt4j.core.problem.Genotype;
import org.opt4j.operator.copy.Copy;

import de.uka.ipd.sdq.dsexplore.helper.UtilisationHelper;
import de.uka.ipd.sdq.dsexplore.launch.DSEWorkflowConfiguration;
import de.uka.ipd.sdq.dsexplore.opt4j.genotype.DesignDecisionGenotype;
import de.uka.ipd.sdq.dsexplore.opt4j.optimizer.heuristic.operators.AbstractHeuristic;
import de.uka.ipd.sdq.dsexplore.opt4j.optimizer.heuristic.operators.HeuristicCandidate;
import de.uka.ipd.sdq.dsexplore.opt4j.representation.DSEIndividual;
import de.uka.ipd.sdq.dsexplore.opt4j.representation.DSEIndividualBuilder;
import de.uka.ipd.sdq.dsexplore.opt4j.start.Opt4JStarter;
import de.uka.ipd.sdq.pcm.core.entity.Entity;
import de.uka.ipd.sdq.pcm.designdecision.AllocationDegree;
import de.uka.ipd.sdq.pcm.designdecision.Choice;
import de.uka.ipd.sdq.pcm.designdecision.EnumerationChoice;
import de.uka.ipd.sdq.pcm.resourceenvironment.ProcessingResourceSpecification;
import de.uka.ipd.sdq.pcm.resourceenvironment.ResourceContainer;
import de.uka.ipd.sdq.pcm.resultdecorator.resourceenvironmentdecorator.UtilisationResult;
import de.uka.ipd.sdq.pcm.resultdecorator.resourceenvironmentdecorator.impl.ProcessingResourceSpecificationResultImpl;

/**
 * Implements a server expansion heuristic. This heuristic finds a
 * highly utilised resource container, creates a new resource container and reallocates 
 * components to this new resource container
 * @author Tom Beyer
 */
public class ServerExpansionImpl extends AbstractHeuristic {
	/**
	 * Maximum number of components reallocated from selected resource container
	 * to a new resource container
	 */
	private int maxNumberOfReplacements;
	/**
	 * If utilisation is larger than this double it will be considered
	 * a high utilisation
	 */
	private double thresholdHighUtilisation;
	
	/**
	 * @param copy Used to copy genotype
	 * @param individualBuilder Used to build individual
	 */
	public ServerExpansionImpl(Copy<Genotype> copy,
			DSEIndividualBuilder individualBuilder, DSEWorkflowConfiguration configuration) {
		super(copy, individualBuilder, configuration);
		// set config
		setHeuristicWeight(configuration.getServerExpansionWeight());
		maxNumberOfReplacements = configuration.getServerExpansionMaxNumberOfReplacements();
		thresholdHighUtilisation = configuration.getServerExpansionThresholdHighUtilisation();
	}
	
	/**
	 * Check whether there is a resource container that has a 
	 * utilisation >= THRESHOLD_HIGH_UTILISATION and is not null
	 * @param individual
	 * @return
	 */
	private boolean doesMatchHighUtilisation(DSEIndividual individual) {
		UtilisationResult maxUtilisationResult = UtilisationHelper
			.getMaxUtilisationResult(individual);
		return maxUtilisationResult != null && 
			maxUtilisationResult.getNormalisedResourceUtilisation() >= thresholdHighUtilisation;
	}
	
	/**
	 * Check whether there is a resource container that has a 
	 * utilisation >= THRESHOLD_HIGH_UTILISATION and there is a least one
	 * unused resource container (unused means no allocated components)
	 * @param individual
	 * @return
	 */
	public boolean doesMatchPrecondition(DSEIndividual individual) {
		return doesMatchHighUtilisation(individual) && !getUnusedResourceContainers(individual).isEmpty();
	}
	
	/**
	 * Generates collection of candidates by applying the server consolidation heuristic
	 * @param individual Individual which the heuristic should be applied to
	 */
	public Collection<HeuristicCandidate> getHeuristicCandidates(DSEIndividual individual) {
		Collection<HeuristicCandidate> candidates = new ArrayList<HeuristicCandidate>();
		if (doesMatchPrecondition(individual)) {
			/*
			 * 1. Find max utilised resource
			 * 2. copy genotype
			 * 3. Get a new (unused) target resource container
			 * 4. Reallocate components to target resource container
			 */
			// 1. Find max utilised resource
			UtilisationResult maxUtilisationResult = UtilisationHelper.getMaxUtilisationResult(individual);
			// 2. copy genotype
			HeuristicCandidate candidate = individualBuilder.buildCandidate(copy.copy(individual.getGenotype()));
			ProcessingResourceSpecification maxProcessingResourceSpec = ((ProcessingResourceSpecificationResultImpl) maxUtilisationResult)
					.getProcessingresourcespecification();
			ResourceContainer resourceContainer = ((ResourceContainer) maxProcessingResourceSpec.eContainer());
			int numberOfReplacements = 0;
			// 3. Get a new (unused) target resource container
			ResourceContainer targetResourceContainer = getRandomUnusedResourceContainer(individual);
			// Iterate through AllocationDegrees and reallocate components
			for (Choice choice : candidate.getGenotype()) {
				if (choice instanceof EnumerationChoice) {
					EnumerationChoice enumerationChoice = (EnumerationChoice)choice;
					if (enumerationChoice.getDegreeOfFreedom() instanceof AllocationDegree) {
						if (enumerationChoice.getEntity() == resourceContainer && numberOfReplacements < maxNumberOfReplacements) {
							// 4. Reallocate components to target resource container
							enumerationChoice.setEntity(targetResourceContainer);
							numberOfReplacements++;
						}
					}
				}
			}
			candidate.setCandidateWeight(getCandidateWeight(maxUtilisationResult));
			candidate.setHeuristic(this);
			candidates.add(candidate);
			increaseCounterOfGeneratedCandidates();
		}
		return candidates;
	}
	

	/**
	 * Calculates weight based on the following scheme: if utilisation equals THRESHOLD_HIGH_UTLISATION
	 * then it will return 0, if utilisation equals 1 it will return 1. Values in between are linearly 
	 * extrapolated. Return values will always be >= 0.
	 * @param utilisationResult
	 * @return Weight based on utilisationResult's utilisation
	 */
	private double getCandidateWeight(UtilisationResult utilisationResult) {
		return Math.max(utilisationResult.getNormalisedResourceUtilisation()/(1.0 - thresholdHighUtilisation) - thresholdHighUtilisation/(1.0 - thresholdHighUtilisation),
				0);
	}
	
	/**
	 * Get a random resource containers that has no allocated components
	 * @param individual
	 * @return
	 */
	private ResourceContainer getRandomUnusedResourceContainer(DSEIndividual individual) {
		Random random = new Random();
		Collection<ResourceContainer> resourceContainers = getUnusedResourceContainers(individual);
		int randomInt = random.nextInt(resourceContainers.size());
		int n = 0;
		for (ResourceContainer container : resourceContainers ) {
			if (n == randomInt) {
				return container;
			}
			n++;
		}
		// wont be executed unless resourceContainers is empty
		return null;
	}
	
	/**
	 * Get all resource containers that have no components
	 * @param individual
	 * @return
	 */
	private Collection<ResourceContainer> getUnusedResourceContainers(DSEIndividual individual) {
		Collection<ResourceContainer> unusedResourceContainers = new HashSet<ResourceContainer>();
		DesignDecisionGenotype genotype = individual.getGenotype();
		// add all resource containers that can be found in the degrees of freedom
		for (Choice choice : genotype) {
			if (choice instanceof EnumerationChoice) {
				EnumerationChoice enumerationChoice = (EnumerationChoice) choice;
				if (enumerationChoice.getDegreeOfFreedom() instanceof AllocationDegree) {
					AllocationDegree allocationDegree = (AllocationDegree) enumerationChoice.getDegreeOfFreedom();
					for (Entity entity : allocationDegree.getDomainOfEntities()) {
						if (entity instanceof ResourceContainer) {
							unusedResourceContainers.add((ResourceContainer)entity);
						}
					}
				}
			}
		}
		// remove all resource containers that are used
		for (Choice choice : genotype) {
			if (choice instanceof EnumerationChoice) {
				EnumerationChoice enumerationChoice = (EnumerationChoice) choice;
				unusedResourceContainers.remove(enumerationChoice.getEntity());
			}
		}
		return unusedResourceContainers;
	}

}
