package de.uka.ipd.sdq.dsexplore.opt4j.representation;


import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.Status;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;

import de.uka.ipd.sdq.dsexplore.designdecisions.alternativecomponents.AlternativeComponent;
import de.uka.ipd.sdq.dsexplore.exception.ChoiceOutOfBoundsException;
import de.uka.ipd.sdq.dsexplore.helper.DegreeOfFreedomHelper;
import de.uka.ipd.sdq.dsexplore.helper.EMFHelper;
import de.uka.ipd.sdq.dsexplore.helper.FixDesignDecisionReferenceSwitch;
import de.uka.ipd.sdq.dsexplore.launch.DSEWorkflowConfiguration;
import de.uka.ipd.sdq.dsexplore.opt4j.genotype.DesignDecisionGenotype;
import de.uka.ipd.sdq.pcm.allocation.AllocationContext;
import de.uka.ipd.sdq.pcm.core.composition.AssemblyContext;
import de.uka.ipd.sdq.pcm.core.entity.Entity;
import de.uka.ipd.sdq.pcm.cost.helper.CostUtil;
import de.uka.ipd.sdq.pcm.designdecision.AllocationDegree;
import de.uka.ipd.sdq.pcm.designdecision.AssembledComponentDegree;
import de.uka.ipd.sdq.pcm.designdecision.ContinousRangeChoice;
import de.uka.ipd.sdq.pcm.designdecision.ContinuousRangeDegree;
import de.uka.ipd.sdq.pcm.designdecision.DegreeOfFreedom;
import de.uka.ipd.sdq.pcm.designdecision.EnumerationChoice;
import de.uka.ipd.sdq.pcm.designdecision.EnumerationDegree;
import de.uka.ipd.sdq.pcm.designdecision.Problem;
import de.uka.ipd.sdq.pcm.designdecision.ProcessingRateDegree;
import de.uka.ipd.sdq.pcm.designdecision.designdecisionFactory;
import de.uka.ipd.sdq.pcm.designdecision.impl.designdecisionFactoryImpl;
import de.uka.ipd.sdq.pcm.repository.RepositoryComponent;
import de.uka.ipd.sdq.pcm.resourceenvironment.ProcessingResourceSpecification;
import de.uka.ipd.sdq.pcm.resourceenvironment.ResourceContainer;
import de.uka.ipd.sdq.pcm.resourcetype.ProcessingResourceType;
import de.uka.ipd.sdq.pcmsolver.models.PCMInstance;

/**
 * The {@link DSEProblem} defines the problem. Therefore, it reads in the 
 * current PCM instances and identifies the available design options to define 
 * the design space. 
 * 
 * @author Anne
 *
 */
public class DSEProblem {
	
	/**
	 * Is changed during the evaluation, as the decisions refer to it.
	 */
	private PCMInstance initialInstance;
	
	private Problem pcmProblem;
	private designdecisionFactory designDecisionFactory;
	

	private List<DesignDecisionGenotype> initialGenotypeList = null;

	private DSEWorkflowConfiguration dseConfig;

	
	/**
	 * @param pcmInstance
	 * @throws CoreException 
	 */
	public DSEProblem(DSEWorkflowConfiguration dseConfig, PCMInstance pcmInstance) throws CoreException{
		this.dseConfig = dseConfig;
		
		boolean newProblem = dseConfig.isNewProblem();
		this.initialInstance = pcmInstance;
		
		this.designDecisionFactory = designdecisionFactoryImpl.init();
		
		if (newProblem){
			initialiseProblem();
		} else {
			//Alternative Component has to be called to create the mapping for all AssemblyContexts and their roles, 
			//in order to be able to exchange implementations later. The design decisions calculated here are not kept, only
			//the mapping inside AlternativeComponent.getInstance()
			AlternativeComponent.getInstance().generateDesignDecisions(pcmInstance);
			Problem problem = loadProblem();
			this.pcmProblem = problem;
			this.initialGenotypeList = determineInitialGenotype(problem);
		}
		
		//TODO: mapping of design decisions to bounds. 
		/*
		 * Meta model design decisions?? Each design decision would know its
		 * bounds and the annotated element or more specific the annotated
		 * property of this element that is exactly the value range than the
		 * bound of the decision. 
		 * 
		 * But what about the genotype? With a ordered list
		 * of design decisions, I can map the genotype entry to a design
		 * decisions. Better group in integer and double, then. -> require CompositeGenotype
		 * 
		 * Also meta-model the genotype as a choice within the range.
		 */		
	}
	
	private Problem loadProblem() throws CoreException {
		String filename = this.dseConfig.getDesignDecisionFileName();
		return this.loadProblem(filename);
	}
	
	private Problem loadProblem(String filename) throws CoreException{
		EObject eproblem = EMFHelper.loadFromXMIFile(filename);
		if (!(eproblem instanceof Problem)){
			throw new CoreException(new Status(Status.ERROR, "de.uka.ipd.sdq.dsexplore", 0, "Cannot read design decision file "+filename+". Please create a new one.", null));
		}
		Problem problem = (Problem)eproblem;
		//Adjust references with the right loaded model objects in memory?

		FixDesignDecisionReferenceSwitch visitor = new FixDesignDecisionReferenceSwitch(this.initialInstance);
		visitor.doSwitch(problem);
		 
		return problem;
	}


	private List<DesignDecisionGenotype> determineInitialGenotype(Problem problem) {
		DesignDecisionGenotype genotype = new DesignDecisionGenotype();
		
		for (DegreeOfFreedom dd : problem.getDesigndecision()) {
			if (dd instanceof EnumerationDegree){

				EnumerationChoice choice = this.designDecisionFactory.createEnumerationChoice();
				choice.setDegreeOfFreedom(dd);

				if (dd instanceof AssembledComponentDegree){
					AssembledComponentDegree acd = (AssembledComponentDegree)dd;
					AssemblyContext ac = (AssemblyContext)acd.getChangeableEntity();
					RepositoryComponent rc = ac.getEncapsulatedComponent_AssemblyContext();
					choice.setEntity(rc);
				} else if (dd instanceof AllocationDegree){
					AllocationDegree ad = (AllocationDegree)dd;
					AllocationContext ac = (AllocationContext)ad.getChangeableEntity();
					ResourceContainer rc = ac.getResourceContainer_AllocationContext();
					choice.setEntity(rc);
				} else throwUnknownDegreeException(dd);
				
				//check if entity is in the domain
				if (!((EnumerationDegree) dd).getDomainOfEntities().contains(choice.getEntity())){
					throw new ChoiceOutOfBoundsException(choice, "Error when determining initial genotype"); 
				}
				
				genotype.add(choice);
			} else if (dd instanceof ContinuousRangeDegree){
				
				ContinousRangeChoice choice = this.designDecisionFactory.createContinousRangeChoice();
				choice.setDegreeOfFreedom(dd);
				
				if (dd instanceof ProcessingRateDegree){
					ProcessingRateDegree prd = (ProcessingRateDegree)dd;
					ResourceContainer rc = (ResourceContainer)prd.getChangeableEntity();
					List<ProcessingResourceSpecification> prsList = rc.getActiveResourceSpecifications_ResourceContainer();
					ProcessingResourceType prt = prd.getProcessingresourcetype();

					ProcessingResourceSpecification rightPrs = null;
					for (ProcessingResourceSpecification prs : prsList) {
						if (EMFHelper.checkIdentity(prs.getActiveResourceType_ActiveResourceSpecification(),prt)){
							rightPrs = prs;
							break;
						}
					}

					if (rightPrs != null){
						double rate = CostUtil.getDoubleFromSpecification(rightPrs.getProcessingRate_ProcessingResourceSpecification().getSpecification());
						choice.setChosenValue(rate);
					} else 
						throw new RuntimeException("Invalid degree of freedom "+dd.toString()+". The references ProcessingResourceType is not available in the given ResourceContainer.");
				} else throwUnknownDegreeException(dd);
				
				genotype.add(choice);
			} else throwUnknownDegreeException(dd);
		};

		//determineProcessingRateDecisions(new ArrayList<DesignDecision>(), genotype);
		//determineAssembledComponentsDecisions(new ArrayList<DesignDecision>(), genotype);
		//determineAllocationDecisions(new ArrayList<DesignDecision>(), genotype);
		List<DesignDecisionGenotype> result = new ArrayList<DesignDecisionGenotype>();
		result.add(genotype);
		return result;
	}


	private void throwUnknownDegreeException(DegreeOfFreedom dd) {
		throw new RuntimeException("Unknown degree of freedom "+dd.toString()+".");		
	}

	/**
	 * Initialises the degrees of freedoms and at the same time determines the initial genotype. 
	 */
	private void initialiseProblem() {
		this.pcmProblem = this.designDecisionFactory.createProblem();
		List<DegreeOfFreedom> dds = this.pcmProblem.getDesigndecision();
		//analyse PCM Instance and create design decisions
		//TODO: could this be possible with a M2M transformation? 
		//First, only get design decisions for making resources faster. 
		this.initialGenotypeList = new ArrayList<DesignDecisionGenotype>();
		DesignDecisionGenotype initialCandidate = new DesignDecisionGenotype();
		determineProcessingRateDecisions(dds,initialCandidate);
		//find equivalent components
		determineAssembledComponentsDecisions(dds,initialCandidate);
		determineAllocationDecisions(dds,initialCandidate);
		//Quickfix: Add a Soap or RMI decision. This is not meta modelled. 
		//determineSOAPOrRMIDecisions();
		
		//TODO: Check if the initial genotype is actually a valid genotype? 
		//(this may not be the case if the degrees of freedom have been reduced for the optimisation?)
		
		
		this.initialGenotypeList.add(initialCandidate);

	}




	/**
	 * XXX: This is not meta modelled and just a quick fix
	 * 
	 */
//	private void determineSOAPOrRMIDecisions(List<DesignDecision> dds, DesignDecisionGenotype genotype) {
//		
//		Configuration connectorConfig = this.initialInstance.getConnectorConfig();
//		FeatureConfig featureConfig = connectorConfig.getDefaultConfig();
//		
//		ConnectorConfigDecision cd = this.designDecisionFactory.createConnectorConfigDecision();
//		SoapOrRmi domain = this.designDecisionFactory.createSoapOrRmi();
//		
//		cd.setDomain(domain);
//		cd.setFeatureconfig(featureConfig);
//		
//		dds.add(cd);
//		
//		//0.0 stands for SOAP, 1.0 for RMI
//		genotype.add(0.0);
//		
//	}

	//TODO: change this to two visitors that either create the design decision and initial genotype or just initial genotype.  
	private void determineAllocationDecisions(List<DegreeOfFreedom> dds, DesignDecisionGenotype genotype) {
		List<AllocationContext> acs = this.initialInstance.getAllocation().getAllocationContexts_Allocation();
		List<ResourceContainer> rcs = this.initialInstance.getResourceEnvironment().getResourceContainer_ResourceEnvironment();
		
		
		
		//each allocation context could be allocated on each container.
		for (AllocationContext ac : acs) {
			AllocationDegree ad = this.designDecisionFactory.createAllocationDegree();
			ad.setChangeableEntity(ac);
			ad.getDomainOfEntities().addAll(rcs);
			dds.add(ad);
			
			EnumerationChoice choice = this.designDecisionFactory.createEnumerationChoice();
			choice.setDegreeOfFreedom(ad);
			choice.setEntity(ac.getResourceContainer_AllocationContext());
			
			genotype.add(choice);
		}
		
		
		
	}


	/**
	 * Be sure to add one design decision and one gene in the initial genotype at once. The index is important.
	 * @param genotypeIndex
	 */
	private void determineAssembledComponentsDecisions(List<DegreeOfFreedom> dds, DesignDecisionGenotype genotype) {
		AlternativeComponent ac = AlternativeComponent.getInstance();
		List<AssembledComponentDegree> decisions = ac.generateDesignDecisions(this.initialInstance);
		
		for (AssembledComponentDegree designDecision : decisions) {
			dds.add(designDecision);
			RepositoryComponent currentlyAssembledComponent = ((AssemblyContext)designDecision.getChangeableEntity()).getEncapsulatedComponent_AssemblyContext();
			
			EnumerationChoice choice = this.designDecisionFactory.createEnumerationChoice();
			choice.setDegreeOfFreedom(designDecision);
			choice.setEntity(currentlyAssembledComponent);
			genotype.add(choice);
			
		}

	}

	/**
	 * Be sure to add one design decision and one gene in the initial genotype at once. The index is important.
	 */
	private void determineProcessingRateDecisions(List<DegreeOfFreedom> dds, DesignDecisionGenotype genotype) {
		List<ResourceContainer> resourceContainers = this.initialInstance.getResourceEnvironment().getResourceContainer_ResourceEnvironment();
		for (ResourceContainer resourceContainer : resourceContainers) {
			List<ProcessingResourceSpecification> resources = resourceContainer.getActiveResourceSpecifications_ResourceContainer();
			for (ProcessingResourceSpecification resource : resources) {
				ProcessingRateDegree decision = this.designDecisionFactory.createProcessingRateDegree();
				decision.setLowerBoundIncluded(false);
				double currentRate = CostUtil.getDoubleFromSpecification(resource.getProcessingRate_ProcessingResourceSpecification().getSpecification());
				//XXX initial assumption: the highest possible processingRate is 10 times the current one.
				decision.setTo(currentRate * 2.0);
				decision.setFrom(currentRate * 0.5);
				decision.setChangeableEntity(resourceContainer);
				decision.setProcessingresourcetype(resource.getActiveResourceType_ActiveResourceSpecification());
				dds.add(decision);
				
				ContinousRangeChoice choice = this.designDecisionFactory.createContinousRangeChoice();
				choice.setDegreeOfFreedom(decision);
				choice.setChosenValue(currentRate);
				
				genotype.add(choice);
				;
			}
		}
	}


	protected DegreeOfFreedom getDesignDecision(int index){
		return this.pcmProblem.getDesigndecision().get(index);
	}

	public List<DegreeOfFreedom> getDesignDecisions(){
		return this.pcmProblem.getDesigndecision();
	}


	public PCMInstance getInitialInstance() {
		return this.initialInstance;
	}
	
	protected List<DesignDecisionGenotype> getInitialGenotypeList(){
		return this.initialGenotypeList;
	}


	public void saveProblem() {

		String filename = this.dseConfig.getDesignDecisionFileName();
		
//		resourceSet.getPackageRegistry().put
//		(designdecisionPackage.eNS_URI, 
//		 designdecisionPackage.eINSTANCE);
		
		EMFHelper.saveToXMIFile(this.pcmProblem, filename);
		
	}

	
	@Override
	public String toString(){
		
		String result = "";
		
		List<DegreeOfFreedom> decisions = this.pcmProblem.getDesigndecision();
		for (DegreeOfFreedom designDecision : decisions) {
			result += DegreeOfFreedomHelper.getDegreeDescription(designDecision)+";";
		}
		
		return result;
		
	}

	
	public void setInitialPopulation(List<DesignDecisionGenotype> population){
		this.initialGenotypeList = population;
	}

	/**
	 * Returns the degree of freedom of the type (or subtype) that has the given 
	 * entity as the changeableEntity.  
	 * @param entity
	 * @param degreeClass
	 * @return The matching DegreeOfFreedom from this problem. 
	 */
	public DegreeOfFreedom getDegree(Entity entity, Class<? extends DegreeOfFreedom> degreeClass){
		List<DegreeOfFreedom> degrees = this.pcmProblem.getDesigndecision();
		for (DegreeOfFreedom degreeOfFreedom : degrees) {
			 
			if (degreeClass.isInstance(degreeOfFreedom) && degreeOfFreedom.getChangeableEntity().equals(entity)){
				return degreeOfFreedom;
			}
		}
		return null;
	}
	
	
}
