package de.uka.ipd.sdq.pcmsolver.transformations.pcm2markov;

import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;
import org.eclipse.emf.common.util.EList;

import de.uka.ipd.sdq.markov.MarkovChain;
import de.uka.ipd.sdq.markov.State;
import de.uka.ipd.sdq.markov.StateType;
import de.uka.ipd.sdq.pcm.repository.ApplicationFailureType;
import de.uka.ipd.sdq.pcm.repository.EnvironmentFailureType;
import de.uka.ipd.sdq.pcm.repository.FailureType;
import de.uka.ipd.sdq.pcm.repository.RepositoryFactory;
import de.uka.ipd.sdq.pcm.resourceenvironment.CommunicationLinkResourceSpecification;
import de.uka.ipd.sdq.pcm.resourcetype.ProcessingResourceType;
import de.uka.ipd.sdq.pcm.seff.AbstractAction;
import de.uka.ipd.sdq.pcm.seff.AbstractBranchTransition;
import de.uka.ipd.sdq.pcm.seff.AcquireAction;
import de.uka.ipd.sdq.pcm.seff.BranchAction;
import de.uka.ipd.sdq.pcm.seff.CollectionIteratorAction;
import de.uka.ipd.sdq.pcm.seff.ExternalCallAction;
import de.uka.ipd.sdq.pcm.seff.FailureHandlingEntity;
import de.uka.ipd.sdq.pcm.seff.FailureOccurrenceDescription;
import de.uka.ipd.sdq.pcm.seff.InternalAction;
import de.uka.ipd.sdq.pcm.seff.LoopAction;
import de.uka.ipd.sdq.pcm.seff.RecoveryBlockAction;
import de.uka.ipd.sdq.pcm.seff.RecoveryBlockAlternativeBehaviour;
import de.uka.ipd.sdq.pcm.seff.ReleaseAction;
import de.uka.ipd.sdq.pcm.seff.ResourceDemandingBehaviour;
import de.uka.ipd.sdq.pcm.seff.ResourceDemandingSEFF;
import de.uka.ipd.sdq.pcm.seff.SeffFactory;
import de.uka.ipd.sdq.pcm.seff.ServiceEffectSpecification;
import de.uka.ipd.sdq.pcm.seff.SetVariableAction;
import de.uka.ipd.sdq.pcm.seff.StartAction;
import de.uka.ipd.sdq.pcm.seff.StopAction;
import de.uka.ipd.sdq.pcm.seff.performance.ParametricResourceDemand;
import de.uka.ipd.sdq.pcm.seff.util.SeffSwitch;
import de.uka.ipd.sdq.pcmsolver.transformations.ContextWrapper;
import de.uka.ipd.sdq.pcmsolver.visitors.EMFHelper;
import de.uka.ipd.sdq.probfunction.math.ManagedPMF;

/**
 * This class represents a visitor for an RDSEFF within a PCM instance. The
 * visitor is used in the transformation from PCM with solved dependencies into
 * a Markov Chain Model for reliability prediction.
 * 
 * @author brosch
 */
public class MarkovSeffVisitor extends SeffSwitch<MarkovChain> {

	/**
	 * A logger to give detailed information about the PCM instance traversal.
	 */
	private static Logger logger = Logger.getLogger(MarkovSeffVisitor.class
			.getName());

	/**
	 * Indicates if the Markov Chain reduction is performed during the
	 * transformation. If so, then the chain as a whole never exists, because
	 * during construction, it is already reduced again.
	 */
	private boolean optimize;

	/**
	 * The ContextWrapper provides easy access to the decorations of the solved
	 * PCM instance.
	 */
	private ContextWrapper contextWrapper;

	/**
	 * The Markov Builder is used to create Markov Chain instances.
	 */
	private MarkovBuilder markovBuilder = new MarkovBuilder();

	private PCMInformationProvider pcmInfoProvider;
	
	/**
	 * The constructor.
	 * 
	 * @param wrapper
	 *            the ContextWrapper provides easy access to the decorations of
	 *            the solved PCM instance
	 * @param descriptors
	 *            the list of resource descriptors
	 * @param optimize
	 *            controls if Markov Chain reduction is performed during
	 *            transformation
	 */
	public MarkovSeffVisitor(final PCMInformationProvider pcmInfoProvider,
							final ContextWrapper wrapper, 
			final boolean optimize) {
		this.pcmInfoProvider = pcmInfoProvider;
		this.contextWrapper = wrapper;
		this.optimize = optimize;
	}

	/**
	 * In the case of a ResourceDemandingBehaviour, the Chain of Actions is
	 * traversed to build an associated Markov Chain.
	 * 
	 * @param behaviour
	 *            the ResourceDemandingBehaviour
	 * @return the resulting Markov Chain
	 */
	@Override
	public MarkovChain caseResourceDemandingBehaviour(
			final ResourceDemandingBehaviour behaviour) {

		// Do the logging:
		logger.debug("Visit ResourceDemandingBehaviour");

		// Go through the chain of actions that constitute this Behaviour. Each
		// action is expected to create its own specific Markov Chain:
		ArrayList<AbstractAction> actions = new ArrayList<AbstractAction>();
		ArrayList<MarkovChain> chains = new ArrayList<MarkovChain>();
		AbstractAction action = (StartAction) EMFHelper.getObjectByType(
				behaviour.getSteps_Behaviour(), StartAction.class);
		while (action != null) {
			MarkovChain specificMarkovChain = (MarkovChain) doSwitch(action);
			actions.add(action);
			chains.add(specificMarkovChain);
			action = action.getSuccessor_AbstractAction();
		}

		// Initialize a new aggregate Markov Chain that has one state for each
		// action of the action chain:
		ArrayList<State> states = new ArrayList<State>();
		MarkovChain aggregateMarkovChain = markovBuilder
				.initResourceDemandingBehaviourMarkovChain(
						"ResourceDemandingBehaviour", actions, states);

		// Incorporate the specific Chains into the aggregate Chain:
		for (int i = 0; i < actions.size(); i++) {
			markovBuilder.incorporateMarkovChain(aggregateMarkovChain, chains
					.get(i), states.get(i), optimize);
		}

		// Return the resulting Markov Chain:
		return aggregateMarkovChain;
	}

	/**
	 * All Actions in the ResourceDemandingSEFF are traversed to build an
	 * associated Markov Chain.
	 * 
	 * @param seff
	 *            the ResourceDemandingSEFF
	 * @return the resulting Markov Chain
	 */
	@Override
	public MarkovChain caseResourceDemandingSEFF(
			final ResourceDemandingSEFF seff) {

		// Do the logging:
		logger.debug("Visit ResourceDemandingSEFF ["
				+ seff.getDescribedService__SEFF().getServiceName() + "]");

		// Go through the chain of actions that constitute this RDSEFF. Each
		// action is expected to create its own specific Markov Chain:
		ArrayList<AbstractAction> actions = new ArrayList<AbstractAction>();
		ArrayList<MarkovChain> chains = new ArrayList<MarkovChain>();
		AbstractAction action = (StartAction) EMFHelper.getObjectByType(seff
				.getSteps_Behaviour(), StartAction.class);
		while (action != null) {
			MarkovChain specificMarkovChain = (MarkovChain) doSwitch(action);
			actions.add(action);
			chains.add(specificMarkovChain);
			action = action.getSuccessor_AbstractAction();
		}

		// Initialize a new aggregate Markov Chain that has one state for each
		// action of the action chain:
		ArrayList<State> states = new ArrayList<State>();
		MarkovChain aggregateMarkovChain = markovBuilder
				.initResourceDemandingBehaviourMarkovChain(seff
						.getDescribedService__SEFF().getServiceName(), actions,
						states);

		// Incorporate the specific Chains into the aggregate Chain:
		for (int i = 0; i < actions.size(); i++) {
			markovBuilder.incorporateMarkovChain(aggregateMarkovChain, chains
					.get(i), states.get(i), optimize);
		}

		// Return the resulting Markov Chain:
		return aggregateMarkovChain;
	}

	/**
	 * A StartAction returns a trivial Markov Chain.
	 * 
	 * @param startAction
	 *            the StartAction
	 * @return the resulting trivial Markov Chain
	 */
	@Override
	public MarkovChain caseStartAction(final StartAction startAction) {

		// Do the logging:
		logger.debug("Visit StartAction [" + startAction.getEntityName() + "]");

		return markovBuilder.initNewMarkovChain(startAction.getEntityName());
	}

	/**
	 * An InternalAction returns a Markov Chain with two Transitions. The first
	 * transition represents Success and the second one Failure of the
	 * InternalAction.
	 * 
	 * @param internalAction
	 *            the InternalAction
	 * @return the resulting Markov Chain
	 */
	@Override
	public MarkovChain caseInternalAction(final InternalAction internalAction) {

		// Do the logging:
		logger.debug("Visit InternalAction [" + internalAction.getEntityName()
				+ "]");

		List<FailureOccurrenceDescription> failureDescriptions=new ArrayList<FailureOccurrenceDescription>();

		// Check if any of the processing resources used by this internal action
		// is currently broken down:
		for (ParametricResourceDemand demand : internalAction
				.getResourceDemand_Action()) {

			// Special case: ignore resource demands of type
			// "SystemExternalResource", as they have been internally introduced
			// by the dependency solver:
			if (demand.getRequiredResource_ParametricResourceDemand()
					.getEntityName().equals("SystemExternalResource")) {
				continue;
			}

			// Get the descriptor that corresponds to the demand:
			ProcessingResourceDescriptor descriptor = pcmInfoProvider.getDescriptor(demand, contextWrapper);
			if (descriptor == null) {
				logger.error("Missing resource description for "
						+ demand.getRequiredResource_ParametricResourceDemand()
								.getEntityName()
						+ " resource demand. Assume resource state = OK.");
				continue;
			}

			// Get the state of the resource:
			ProcessingResourceState state = descriptor.getCurrentState();
			if (state == null) {
				logger.error("Resource state no set for "
						+ demand.getRequiredResource_ParametricResourceDemand()
								.getEntityName()
						+ " resource demand. Assume resource state = OK.");
				continue;
			}

			// If the resource is broken down, the internal action fails with
			// 100% probability:
			if (state.equals(ProcessingResourceState.NA)) {
				failureDescriptions.add(createOccurrenceDescriptionForResource(
						demand.getRequiredResource_ParametricResourceDemand()));
				break;
			}
		}

		//if a resource failure was added, the failure probability is set to 100%, so the other failures do not matter.
		if(failureDescriptions.isEmpty()) {
			//add all remaining failure probabilities
			failureDescriptions.addAll(internalAction.getFailureOccurrenceDescriptions());
		}
		// Build the corresponding Markov Chain:
		return markovBuilder.initInternalMarkovChain(internalAction
				.getEntityName(), failureDescriptions);
	}

	/**
	 * Creates a FalilureOccurrenceDescription for an environment failure. 
	 * That is, the FailureType is EnvironmentFailureType referencing the specified resource type
	 * and a failure probability of 1.0.
	 * 
	 * @param resourceType
	 *            the resource type to create the failure occurrence for
	 * @return a corresponding FailureOccurrenceDescription. Note that this FailureOccurrenceDescription
	 * is "virtual", i.e not part of the model.
	 */
	private FailureOccurrenceDescription createOccurrenceDescriptionForResource(ProcessingResourceType resourceType){
		FailureOccurrenceDescription failureOccurrence=SeffFactory.eINSTANCE.createFailureOccurrenceDescription();
		failureOccurrence.setFailureProbability(1.0);
		
		EnvironmentFailureType failureType= RepositoryFactory.eINSTANCE.createEnvironmentFailureType();
		//set the name to the resource name
		failureType.setEntityName(resourceType.getEntityName());
		failureType.setProcessingresourcetype(resourceType);
		failureOccurrence.setFailureType(failureType);
		
		return failureOccurrence;
	}
	
	/**
	 * For a LoopAction, first the Markov Chain of the body behaviour is built.
	 * The result is then inserted into a new Markov Chain that has one State
	 * for each of the possible iteration counts of the loop.
	 * 
	 * @param loopAction
	 *            the LoopAction
	 * @return the resulting Markov Chain
	 */
	@Override
	public MarkovChain caseLoopAction(final LoopAction loopAction) {

		// Do the logging:
		logger.debug("Visit LoopAction [" + loopAction.getEntityName() + "]");

		// Determine the inner Markov Chain associated with the loop behaviour:
		MarkovChain specificMarkovChain = (MarkovChain) doSwitch(loopAction
				.getBodyBehaviour_Loop());

		// Get the solved loop probability mass function:
		ManagedPMF pmf = contextWrapper.getLoopIterations(loopAction);

		// Initialize the aggregate Markov Chain representing the loop:
		MarkovChain aggregateMarkovChain = markovBuilder.initLoopMarkovChain(
				loopAction.getEntityName(), pmf);

		// Incorporate the specific MarkovChain into the aggregate one:
		ArrayList<State> statesToReplace = new ArrayList<State>();
		for (int i = 0; i < aggregateMarkovChain.getStates().size(); i++) {
			if (aggregateMarkovChain.getStates().get(i).getType().equals(
					StateType.DEFAULT)) {
				statesToReplace.add(aggregateMarkovChain.getStates().get(i));
			}
		}
		for (int i = 0; i < statesToReplace.size(); i++) {
			markovBuilder.incorporateMarkovChain(aggregateMarkovChain,
					specificMarkovChain, statesToReplace.get(i), optimize);
		}

		// Return the result:
		return aggregateMarkovChain;
	}

	/**
	 * A collection iterator action is handled the same way as an ordinary loop.
	 * 
	 * @param collectionIteratorAction
	 *            the CollectionIteratoraction
	 * @return the resulting Markov Chain
	 */
	@Override
	public MarkovChain caseCollectionIteratorAction(
			final CollectionIteratorAction collectionIteratorAction) {

		// Do the logging:
		logger.debug("Visit CollectionIteratorAction ["
				+ collectionIteratorAction.getEntityName() + "]");

		// Determine the inner Markov Chain associated with the loop behaviour:
		MarkovChain specificMarkovChain = (MarkovChain) doSwitch(collectionIteratorAction
				.getBodyBehaviour_Loop());

		// Get the solved loop probability mass function:
		ManagedPMF pmf = contextWrapper
				.getLoopIterations(collectionIteratorAction);

		// Initialize the aggregate Markov Chain representing the loop:
		MarkovChain aggregateMarkovChain = markovBuilder.initLoopMarkovChain(
				collectionIteratorAction.getEntityName(), pmf);

		// Incorporate the specific MarkovChain into the aggregate one:
		ArrayList<State> statesToReplace = new ArrayList<State>();
		for (int i = 0; i < aggregateMarkovChain.getStates().size(); i++) {
			if (aggregateMarkovChain.getStates().get(i).getType().equals(
					StateType.DEFAULT)) {
				statesToReplace.add(aggregateMarkovChain.getStates().get(i));
			}
		}
		for (int i = 0; i < statesToReplace.size(); i++) {
			markovBuilder.incorporateMarkovChain(aggregateMarkovChain,
					specificMarkovChain, statesToReplace.get(i), optimize);
		}

		// Return the result:
		return aggregateMarkovChain;
	}

	/**
	 * For a BranchAction, first the Markov Chain of each of the transition
	 * behaviours is built. The results are then inserted into a new Markov
	 * Chain that has one State for each of the possible branches.
	 * 
	 * @param branchAction
	 *            the BranchAction
	 * @return the resulting Markov Chain
	 */
	@Override
	public MarkovChain caseBranchAction(final BranchAction branchAction) {

		// Do the logging:
		logger.debug("Visit BranchAction [" + branchAction.getEntityName()
				+ "]");

		// Determine the inner Markov Chains associated with the branch
		// behaviours:
		EList<AbstractBranchTransition> transitions = branchAction
				.getBranches_Branch();
		ArrayList<MarkovChain> specificMarkovChains = new ArrayList<MarkovChain>();
		ArrayList<Double> branchProbabilities = new ArrayList<Double>();
		for (int i = 0; i < transitions.size(); i++) {
			branchProbabilities.add(contextWrapper
					.getBranchProbability(transitions.get(i)));
			specificMarkovChains.add((MarkovChain) doSwitch(transitions.get(i)
					.getBranchBehaviour_BranchTransition()));
		}

		// Initialize the aggregate Markov Chain representing the loop:
		MarkovChain aggregateMarkovChain = markovBuilder.initBranchMarkovChain(
				branchAction.getEntityName(), branchProbabilities);

		// Incorporate the specific MarkovChain into the aggregate one:
		ArrayList<State> statesToReplace = new ArrayList<State>();
		for (int i = 0; i < aggregateMarkovChain.getStates().size(); i++) {
			if (aggregateMarkovChain.getStates().get(i).getType().equals(
					StateType.DEFAULT)) {
				statesToReplace.add(aggregateMarkovChain.getStates().get(i));
			}
		}
		for (int i = 0; i < statesToReplace.size(); i++) {
			markovBuilder.incorporateMarkovChain(aggregateMarkovChain,
					specificMarkovChains.get(i), statesToReplace.get(i),
					optimize);
		}

		// Return the result:
		return aggregateMarkovChain;
	}

	/**
	 * Handles an ExternalCallAction. Generates a Markov Chain of the executing behavior.  
	 * The Markov Chain is generated by {@link MarkovSeffVisitor.createExternalCallActionChain}.
	 * If an retry count is set in the internal action, the chain is repeatedly 
	 * appended to the main chains failure state.
	 * 
	 * @param externalCallAction
	 *            the ExternalCallAction
	 * @return the resulting Markov Chain.
	 */
	@Override
	public MarkovChain caseExternalCallAction(
			final ExternalCallAction externalCallAction) {
		
		if(externalCallAction==null)
			throw new IllegalArgumentException("Passed ExternalCallAction is null.");
		
		// Get a reference to the executing SEFF:
		ServiceEffectSpecification seff = contextWrapper
				.getNextSEFF(externalCallAction);

		// Distinguish the type of the ExternalCall:
		if (seff == null) {

			// Do the logging:
			logger.debug("Visit SystemExternalCallAction ["
					+ externalCallAction.getEntityName() + "]");

			// Get the FailureProbability of the SystemExternalCall (this is, by
			// now, a hack):
			// String failureProbabilityExpression = failProbList.get(sysExList
			// .indexOf(externalCallAction.getId()));

			// A SystemExternalCall is treated like an InternalAction - it has
			// just one fixed failureProbability:
			// return markovBuilder.initInternalMarkovChain(externalCallAction
			// .getEntityName(), failureProbabilityExpression);

			// At the moment, no failure probabilities for system-external calls
			// are supported:
			return markovBuilder.initNewMarkovChain(externalCallAction
					.getEntityName());

		} else {

			//int retryCount = externalCallAction.getRetryCount();
			int retryCount = 0;
			if(retryCount<0)
			{
				logger.info("Retry count is negative. Assuming 0.");
				retryCount=0;
			}
			
			//create a Markov Chain for one call
			MarkovChain chain = createExternalCallActionChain(externalCallAction);
			
			List<String> failureTypes=getFailureTypeNames(externalCallAction);
			
			MarkovChain handlingChain=markovBuilder.copyMarkovChain(chain);
			//if a retry count is specified, append the chain to itself as a failure handling chain at the failure state
			//this models the re-execution
			for (int i = 0; i < retryCount; ++i) {
				markovBuilder.appendFailureHandlingChain(chain, handlingChain, failureTypes, optimize);
			}

			return chain;
		}
	}

	private List<String> getFailureTypeNames(FailureHandlingEntity entity) {
		List<String> failureTypes=new ArrayList<String>();
		for(FailureType failureType : entity.getFailuretype()) {
			if(failureType instanceof ApplicationFailureType) {
				failureTypes.add(failureType.getEntityName());
			}
			else if(failureType instanceof EnvironmentFailureType) {
				failureTypes.add(((EnvironmentFailureType)failureType).getProcessingresourcetype().getEntityName());
			}
			else
				throw new MarkovException("Unsupported failure type "+failureType.getClass().getName());
		}
			
		
		return failureTypes;
	}
	/**
	 * An ExternalCallAction returns the Markov Chain of the executing
	 * behaviour, or a trivial Markov Chain if it is a SystemExternalCall.
	 * 
	 * @param externalCallAction
	 *            the ExternalCallAction
	 * @return the resulting Markov Chain.
	 */
	public MarkovChain createExternalCallActionChain(
			final ExternalCallAction externalCallAction) {

		// Get a reference to the executing SEFF:
		ServiceEffectSpecification seff = contextWrapper
				.getNextSEFF(externalCallAction);
		
		// Do the logging:
		logger.debug("Visit ExternalCallAction ["
				+ externalCallAction.getEntityName() + "]");

		// For the new SEFF, we need a new ContextWrapper. As during the
		// creation of the new ContextWrapper, the old one is altered (which
		// is certainly bad programming style!), we need to save a copy of
		// the old one and restore it after generating the new one:
		ContextWrapper originalContextWrapper = (ContextWrapper) contextWrapper
				.clone();
		ContextWrapper newContextWrapper = contextWrapper
				.getContextWrapperFor(externalCallAction);
		contextWrapper = originalContextWrapper;

		// Return the Markov Chain of the executing SEFF:
		MarkovChain innerMarkovChain = new MarkovSeffVisitor(this.pcmInfoProvider, newContextWrapper, 
				optimize).doSwitch(seff);

		// Check if the external call crosses the border of one resource
		// container and uses a communication link:
		CommunicationLinkResourceSpecification commLink = contextWrapper
				.getConcreteLinkingResource(externalCallAction);

		// If a communication link is used, consider the possibility that
		// the call fails:
		if (commLink != null) {

			// The call can be modeled as a behavior with three steps: the
			// sending of the message, the remote execution, and the message
			// return. All of these steps can fail.
			ArrayList<State> states = new ArrayList<State>();
			ArrayList<String> names = new ArrayList<String>();
			names.add("Messaging");
			names.add("Execution");
			names.add("Returning");
			MarkovChain aggregateMarkovChain = markovBuilder
					.initBehaviourMarkovChain(externalCallAction
							.getEntityName(), names, states);

			List<FailureOccurrenceDescription> occurrences=new ArrayList<FailureOccurrenceDescription>();
			FailureOccurrenceDescription description =createOccurrenceDescriptionForResource(commLink.getCommunicationLinkResourceType_CommunicationLinkResourceSpecification());
			description.setFailureProbability(commLink.getFailureProbability());
			occurrences.add(description);
						
			// The first and third Steps can be modeled like an Internal
			// Action which either fails or succeeds:
			MarkovChain messagingMarkovChain = markovBuilder
					.initInternalMarkovChain("Messaging", occurrences);
			MarkovChain returnMarkovChain = markovBuilder
					.initInternalMarkovChain("Returning", occurrences);

			// The second step is the already computed inner Markov Chain.
			// Incorporate all steps into the aggregate chain:
			markovBuilder.incorporateMarkovChain(aggregateMarkovChain,
					messagingMarkovChain, states.get(0), optimize);
			markovBuilder.incorporateMarkovChain(aggregateMarkovChain,
					innerMarkovChain, states.get(1), optimize);
			markovBuilder.incorporateMarkovChain(aggregateMarkovChain,
					returnMarkovChain, states.get(2), optimize);

			// Return the result:
			return aggregateMarkovChain;

		} else {

			// If no communication link is used, then the Markov Chain just
			// has to reflect the inner SEFF behavior:
			return innerMarkovChain;
		}
	}
	
	@Override
	public MarkovChain caseRecoveryBlockAction(final RecoveryBlockAction action) {
		// Do the logging:
		logger.debug("Visit RecoveryBlockAction [" + action.getEntityName() + "]");

		List<RecoveryBlockAlternativeBehaviour> behaviours=action.getRecoveryBlockalternativeBehaviours();

		
		if(behaviours.size()==0)
			throw new MarkovException("RecoveryBlockAction '"+action.getEntityName()+"' does not specify any behaviours.");
		
		RecoveryBlockAlternativeBehaviour previousAlternative=behaviours.get(0);
		MarkovChain theChain = doSwitch(previousAlternative);

		while(previousAlternative.getNextAlternative() !=null) {

			previousAlternative=previousAlternative.getNextAlternative();
			List<String> failureTypes=getFailureTypeNames(previousAlternative);
			
			MarkovChain handlingChain=doSwitch(previousAlternative);
			markovBuilder.appendFailureHandlingChain(theChain, handlingChain, failureTypes, optimize);

		}
		
		return theChain;
	}
	
	/**
	 * An Acquire Action returns a trivial Markov Chain.
	 * 
	 * @param acquireAction
	 *            the acquire action
	 * @return the resulting Markov Chain.
	 * @see de.uka.ipd.sdq.pcm.seff.util.SeffSwitch#caseAcquireAction(de.uka.ipd.sdq.pcm.seff.AcquireAction)
	 */
	@Override
	public MarkovChain caseAcquireAction(final AcquireAction acquireAction) {

		// Do the logging:
		logger.debug("Visit AcquireAction [" + acquireAction.getEntityName()
				+ "]");

		return markovBuilder.initNewMarkovChain(acquireAction.getEntityName());
	}

	/**
	 * A Release Action returns a trivial Markov Chain.
	 * 
	 * @param releaseAction
	 *            the release action
	 * @return the resulting Markov Chain.
	 * @see de.uka.ipd.sdq.pcm.seff.util.SeffSwitch#caseAcquireAction(de.uka.ipd.sdq.pcm.seff.AcquireAction)
	 */
	@Override
	public MarkovChain caseReleaseAction(final ReleaseAction releaseAction) {

		// Do the logging:
		logger.debug("Visit ReleaseAction [" + releaseAction.getEntityName()
				+ "]");

		return markovBuilder.initNewMarkovChain(releaseAction.getEntityName());
	}

	/**
	 * A SetVariableAction returns a trivial Markov Chain.
	 * 
	 * @param setVariableAction
	 *            the SetVariableAction
	 * @return the resulting Markov Chain
	 * @see de.uka.ipd.sdq.pcm.seff.util.SeffSwitch#caseSetVariableAction(de.uka.ipd.sdq.pcm.seff.SetVariableAction)
	 */
	@Override
	public MarkovChain caseSetVariableAction(
			final SetVariableAction setVariableAction) {

		// Do the logging:
		logger.debug("Visit SetVariableAction ["
				+ setVariableAction.getEntityName() + "]");

		return markovBuilder.initNewMarkovChain(setVariableAction
				.getEntityName());
	}

	/**
	 * A StopAction returns a trivial Markov Chain.
	 * 
	 * @param stopAction
	 *            the StopAction
	 * @return the resulting trivial Markov Chain
	 */
	@Override
	public MarkovChain caseStopAction(final StopAction stopAction) {

		// Do the logging:
		logger.debug("Visit StopAction [" + stopAction.getEntityName() + "]");

		return markovBuilder.initNewMarkovChain(stopAction.getEntityName());
	}

		}
